import java.awt.AlphaComposite;
import java.awt.Color;
import java.awt.Composite;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics2D;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;

import javax.imageio.ImageIO;


public class GRFrame  {

	String mapname = null;
	BufferedImage the_map = null;
	float map_height = 0;
	float map_width = 0;
	Graphics2D g2 = null;
	
	static final float map_offset_Y = 84;
	static final float map_offset_X = 28;
	
	Color fill_color;
	
	public GRFrame(int r, int g, int b) 
	{
		fill_color = new Color(r, g, b);
	}
	
	public GRFrame()
	{
		fill_color = new Color(0, 0, 128);
	}
	
	public boolean Load()
	{
		return Load(MPAnalyzer.GetConfig().map_file_name);
	}

	public boolean Load(String mappath)
	{
		mapname = mappath;
		try {
			the_map = ImageIO.read(new File(mappath));
		} catch (IOException e) {
		}
		
		if (the_map == null) {
			return false;
		}
		
		map_height = the_map.getHeight();
		map_width = the_map.getWidth();
		g2 = the_map.createGraphics();
				
		return true;
	}
	
	public boolean Store(String newname)
	{
		// crop it down first
		int x = Math.round(map_width / 400f);
		int width = Math.round(map_width - 2 * x);
		int y = Math.round(map_height / 5);
		int height = Math.round(map_height - 2 * y);
		BufferedImage store_image = the_map.getSubimage(x, y, width, height);

		boolean result = false;
		try {
		    File outputfile = new File(newname);
		    ImageIO.write(store_image, "png", outputfile);
		} catch (IOException e) {
		}

		return result;
	}
	
	public class MapCoord {
		float x;
		float y;
		
		// map from a geo coord to our map surface
		// latitude (height) goes from -90 to +90
		// longitude (width) goes from -180 to +180
		public MapCoord(float c_lat, float c_long)
		{
			
		}
		
	}

    double latitudeToPixelY(double latitude) 
    {
        double sinLatitude = Math.sin(latitude * (Math.PI / 180));
        return (0.5 - Math.log((1 + sinLatitude) / (1 - sinLatitude)) / (4 * Math.PI))
                * ((long) map_height);
    }
    
    int PixelYToMap(double pixely)
    {
    	return (int)Math.round(pixely + map_offset_Y);
    }
    
    double longitudeToPixelX(double longitude) 
    {
        return (longitude + 180) / 360 * map_width;
    }

    int PixelXToMap(double pixelx)
    {
    	return (int)Math.round(pixelx + map_offset_X);
    }
    
	public void TestLatitude()
	{
		for (int i = -90; i <= 90; i += 10)  {
			System.out.println(i + " " + PixelYToMap(latitudeToPixelY(i)));
		}
	}
	
	protected Color percentToColorTestX(float percent)
	{
		percent = Math.max(0f, percent);
		float x = 1.0f / (percent * percent);
		x = Math.min(x / 100f, 1f);
		float h = .5f + percent * .4f;
		float s = .3f + percent * .7f;
		float b = 1f - percent * .5f;
		System.out.println(percent + ": " + h + "," + s + "," + b);
		Color thecolor = Color.getHSBColor(h, s, b);
		return thecolor;
	}

	protected Color percentToColorTestY(float percent)
	{
		percent = (percent - .5f) * 2f;
		percent = Math.min(100f, Math.max(0f, percent));
		float h = .2f + percent * .8f;
		float s = .5f + percent * .5f;
		float b = 1f - (float)percent * .3f;
		// System.out.println(percent + ": " + h + "," + s + "," + b);
		Color thecolor = Color.getHSBColor(h, s, b);
		return thecolor;
	}
	
	// i actually think - 1 / ln(x) is best curve
	
	
	// intra/inter
	protected Color percentToColor(float percent)
	{
		float hpercent = (percent - .51f) /(1f -.51f);
		hpercent = Math.min(1f, Math.max(0f, hpercent));
		float h = .1f + hpercent * .9f;
		float s = .5f + percent * .5f;
		float b = 1f - (float)percent * .3f;
		// System.out.println(percent + ": " + h + "," + s + "," + b);
		Color thecolor = Color.getHSBColor(h, s, b);
		return thecolor;
	}
	
	// tempo
	protected Color percentToColorT(float percent)
	{
		float hpercent = (percent - .51f) /(1f -.51f);
		hpercent = Math.min(1f, Math.max(0f, hpercent));
		float h = .2f + percent * .5f;
		float s = .1f + percent * .9f;
		float b = 1f - (float)percent * .2f;

		// float h = .1f + hpercent * .9f;
		// float s = .5f + percent * .5f;
		// float b = 1f - (float)percent * .3f;
		// System.out.println(percent + ": " + h + "," + s + "," + b);
		Color thecolor = Color.getHSBColor(h, s, b);
		return thecolor;
	}
	
	// expert
	protected Color percentToColorE(float percent)
	{
		float hpercent = (percent - .51f) /(1f -.51f);
		hpercent = Math.min(1f, Math.max(0f, hpercent));
		float h = .3f + percent * .5f;
		float s = .1f + percent * .9f;
		float b = 1f - (float)percent * .2f;

		// float h = .1f + hpercent * .9f;
		// float s = .5f + percent * .5f;
		// float b = 1f - (float)percent * .3f;
		// System.out.println(percent + ": " + h + "," + s + "," + b);
		Color thecolor = Color.getHSBColor(h, s, b);
		return thecolor;
	}

	private float percentToComposite(float percent)
	{
		float f = Math.min(1.0f, Math.max(0.0f, percent));
		return f;
	}
	
	private float radiusToComposite(double radius)
	{
		double f = .3; // + ((3 - radius) / 3) * .2;
		f = Math.min(Math.max(f, 0), 1);
		return (float)f;
	}
	
	public void DrawPoint(String point, double radius, float percent)
	{
	    String[] tokens = point.split(",");
		double latitude = Float.parseFloat(tokens[0]);
		double longitude = Float.parseFloat(tokens[1]);
		
		percent = percentToComposite(percent);
		DrawPoint(latitude, longitude, radius, percentToColor(percent));
	}
	
	public void DrawKeyString(double latitude, double longitude, float percent)
	{
		percent = percentToComposite(percent);
		Color thecolor = percentToColor(percent);
		String s = Float.toString(Math.round(percent * 1000f) / 1000f);
		int x = PixelXToMap(longitudeToPixelX(longitude));
		int y = PixelYToMap(latitudeToPixelY(latitude));
		double deltay = calcCircumferenceAtPoint(latitude, longitude, 1.0).y;
		DrawString(s, x, y + (int)Math.round(deltay / 3), 1.0f, thecolor);
		longitude -= 2f;
		DrawPoint(latitude, longitude, 1.0, thecolor);
	}
	
	public void DrawKey()
	{
		int max = 10;
		int initial_latitude = -10;
		int initial_longitude = -120;

        Font font = new Font("Arial", Font.BOLD, 36);
        g2.setFont(font);
		FontMetrics fm = g2.getFontMetrics();
		String str = "1.0";

		Rectangle2D txtbounds = fm.getStringBounds(str, g2);
		int textWidth = (int)txtbounds.getWidth();
		int textHeight = (int)txtbounds.getHeight();
		
		Rectangle2D.Double bounds = getRectAtPoint(initial_latitude, initial_longitude);		
		Rectangle2D.Double fillrect = new Rectangle2D.Double();
		fillrect.x = bounds.x;
		fillrect.y = bounds.y;
		fillrect.width = textWidth * 2.1f;
		fillrect.height = textHeight * 11;
		
	    Composite c = AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 1f);
	    g2.setComposite(c);
		g2.setPaint(Color.white);
		g2.fill(fillrect);		
		
		int margin = 10;
		fillrect.x += margin + 1;
		fillrect.y += margin * 1.5f;
		fillrect.width -= margin * 2;
		fillrect.height -= margin * 2;	
		
		for (int i = 0; i < max; i += 1) {
			float percent = (float)((max - i) / (float)max);
			percent = percentToComposite(percent);
			Color thecolor = percentToColor(percent);
			String s = Float.toString(Math.round(percent * 100f) / 100f);
			fillrect.width = bounds.width * 2;
			fillrect.height = bounds.height * 2;
			
			c = AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 1f);
		    g2.setComposite(c);
			g2.setPaint(thecolor);
			Ellipse2D.Double e = new Ellipse2D.Double(fillrect.x, fillrect.y, fillrect.width, fillrect.height);
			g2.fill(e);
			g2.drawString(s, (int)(fillrect.x + fillrect.width + 12), (int)(fillrect.y + textHeight / 2));
			
			fillrect.y += textHeight + 4;
		}
	}
	
	private Point2D.Double calcCircumferenceAtPoint(double latitude, double longitude, double radius)
	{
		return calcCircumferenceAtPointX(latitude, longitude, radius);
	}
	
	private Point2D.Double calcCircumferenceAtPointX(double latitude, double longitude, double radius)
	{
		double y_zero = latitudeToPixelY(1) - latitudeToPixelY(0);
		double y_lat = latitudeToPixelY(Math.abs(latitude)) - latitudeToPixelY(Math.abs(latitude) - 1.0);
		double scaley = (y_lat / y_zero) * radius;
		
		// a bit redundant but who knows how longitude params may change
		double x_zero = longitudeToPixelX(1) - longitudeToPixelX(0);
		double x_long = longitudeToPixelX(Math.abs(longitude)) - longitudeToPixelX(Math.abs(longitude) - 1.0);
		double scalex = (x_long / x_zero) * radius;
		
		double regionsX = 360.0f;
		double regionsPerMapX = (map_width ) / regionsX;
		if (regionsPerMapX < 1) {
			System.err.println("calcCircumference:  map too small for region size " + map_width);
		}
		
		double regionsY = 360.0f;
		double regionsPerMapY = (map_height ) / regionsY;
		
		Point2D.Double pt = new Point2D.Double();
		pt.y = regionsPerMapY * scaley * 11f/9f;
		pt.x = regionsPerMapX * scalex * 11f/9f;
		return pt;
	}
	
	public void DrawString(String s, int x, int y, float percent, Color thecolor)
	{
	    Composite c = AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 1f); // percent);
	    g2.setComposite(c);
		g2.setPaint(thecolor);
		g2.drawString(s, x, y);
	}
	
	public void HighlightPoint(double latitude, double longitude, double radius)
	{
		Rectangle2D.Double rect = getRectAtPoint(latitude, longitude, radius);
		Ellipse2D circle = new Ellipse2D.Double(rect.x, rect.y, rect.width, rect.height);
		g2.fill(circle);		
	
	    Composite c = AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 1f);
	    g2.setComposite(c);
		g2.setPaint(Color.black);
	}

	private Rectangle2D.Double getRectAtPoint(double latitude, double longitude)
	{
		return getRectAtPoint(latitude, longitude, 1.0);
	}

	private Rectangle2D.Double getRectAtPoint(double latitude, double longitude, double radius)
	{
		Point2D.Double pt = calcCircumferenceAtPoint(latitude, longitude, radius);
		int x = PixelXToMap(longitudeToPixelX(longitude));
		int y = PixelYToMap(latitudeToPixelY(latitude));
		Rectangle2D.Double rect = new Rectangle2D.Double();
		rect.x = x - pt.x / 2;
		rect.y = y - pt.y / 2;
		rect.width = pt.x;
		rect.height = pt.y;
		return rect;
	}
	
	private void DrawPoint(double latitude, double longitude, double radius, Color thecolor)
	{
		Rectangle2D.Double rect = getRectAtPoint(latitude, longitude, radius);
		Ellipse2D.Double e = new Ellipse2D.Double(rect.x, rect.y, rect.width, rect.height);
	    Composite c = AlphaComposite.getInstance(AlphaComposite.SRC_OVER, radiusToComposite(radius));
	    g2.setComposite(c);
		g2.setPaint(thecolor);
		g2.fill(e);
	}	
	
	public void TestCircles()
	{
		int initial_latitude = -10;
		int initial_longitude = -160;

        Font font = new Font("Arial", Font.BOLD, 36);
        g2.setFont(font);
		FontMetrics fm = g2.getFontMetrics();
		String str = "1.0";

		Rectangle2D txtbounds = fm.getStringBounds(str, g2);
		// int textWidth = (int)txtbounds.getWidth();
		int textHeight = (int)txtbounds.getHeight();
		
		Rectangle2D.Double bounds = getRectAtPoint(initial_latitude, initial_longitude);		
		Rectangle2D.Double fillrect = new Rectangle2D.Double();
		fillrect.x = bounds.x;
		fillrect.y = bounds.y;
		fillrect.width = textHeight * 11;
		fillrect.height = textHeight * 11;
		
	    Composite c = AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 1f);
	    g2.setComposite(c);
		g2.setPaint(Color.white);
		g2.fill(fillrect);		
		
		double orgx = fillrect.x + fillrect.width /2;
		double orgy = fillrect.y+ fillrect.height/2;
		double r = fillrect.width / 2;
		
		Ellipse2D.Double e = new Ellipse2D.Double(orgx - r, orgy - r, r * 2, r * 2);
	    c = AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 1f);
	    g2.setComposite(c);
		g2.setPaint(Color.yellow);
		g2.fill(e);		
		
	    c = AlphaComposite.getInstance(AlphaComposite.SRC_OVER, .2f);
	    g2.setComposite(c);

		double r1 = Math.sqrt(r*r + r*r) / 2;
		double r2 = r / 2;
		
		r = r1;
		g2.setPaint(Color.red);
		double theta = Math.PI / 6;
		for (int i = 0; i < 4; i++) {
			double x = orgx + r * Math.cos(theta) - r;
			double y = orgy + r * Math.sin(theta) - r;
			e = new Ellipse2D.Double(x, y, r * 2, r * 2);
			g2.fill(e);		
			theta += Math.PI / 2;
		}

		r = r1;
		theta = Math.PI / 6;
		for (int i = 8; i < 8; i++) {
			double x = orgx + r * Math.cos(theta) - r;
			double y = orgy + r * Math.sin(theta) - r;
			e = new Ellipse2D.Double(x, y, r * 2, r * 2);
			g2.fill(e);		
			theta += Math.PI / 4;
		}
		
		r = r2;
		g2.setPaint(Color.pink);
		theta = Math.PI / 12;
		for (int i = 16; i < 16; i++) {
			double x = orgx + r * Math.cos(theta) - r;
			double y = orgy + r * Math.sin(theta) - r;
			e = new Ellipse2D.Double(x, y, r * 2, r * 2);
			g2.fill(e);		
			theta += Math.PI / 8;
		}
	}
}

