import java.io.File;
import java.util.ArrayList;
import java.util.Collections;


public class GRMapData {
	String name;
	String datapath;
	String dataname;
	String storename;
	int min_uniques;
	float default_radius;
	
	static final int NAME_INDEX = 0;
	static final int RADIUS_INDEX = 1;
	static final int INTER_INDEX = 2;
	static final int INTRA_INDEX = 3;
	static final int UNIQUES_INDEX = 6;
	static final int URBAN_INDEX = 7;
	static final int EXPERT_INDEX = 8;
	static final int TEMPO_INDEX = 9;
		
	GRFrame grf;
		
	private class GRMapDataItem implements Comparable<GRMapDataItem> {
		private String data_name;
		private double radius;
		private float external_affinity;
		private float internal_affinity;
		private int uniques;
		private float expert;
		private float tempo;
		
		public GRMapDataItem(String name, float r, float ext_aff, float int_aff, int unq, float exp, float tmpo)
		{
			data_name = name;
			radius = r; // / default_radius;
			external_affinity = ext_aff;
			internal_affinity = int_aff;
			uniques = unq;
			expert = exp;
			tempo = tmpo;
		}
		
		public int compareTo(GRMapDataItem item)
		{
			return (int)Math.round((radius - item.radius) * 1000);
		}
	}
	
	private ArrayList<GRMapDataItem> list = null;

	public GRMapData()
	{
		name = MPAnalyzer.GetConfig().the_file_target;
		datapath = "data" + File.separator + "maps" + File.separator;
		dataname = datapath + name + ".xls";
		storename = datapath + name + ".png";
		min_uniques = MPAnalyzer.GetConfig().min_uniques;
	}
	
	public void LoadGraph()
	{
		grf = new GRFrame();
		grf.Load();		
	}
	
	public void StoreGraph()
	{
		grf.Store(storename);
	}
	
	public void GraphPointsX()
	{
		long total = 0;
		float max = 0;
		float min = 1000;
		grf.DrawKey();
		list = loadSegment(dataname);
		for (int i = list.size() - 1; i >= 0; i--) {
			GRMapDataItem item = list.get(i);
			if (item.uniques >= min_uniques) {
				if (item.external_affinity > max) {
					max = item.external_affinity;
				}
				if (item.external_affinity < min) {
					min = item.external_affinity;
				}
			}
		}
		float range = max - min;
		for (int i = list.size() - 1; i >= 0; i--) {
			GRMapDataItem item = list.get(i);
			if (item.uniques >= min_uniques) {
				float t = item.external_affinity;
				// float t = (item.external_affinity - min) / range;
				grf.DrawPoint(item.data_name, item.radius, t);
				total += item.uniques;
			}
		}

		System.out.println("total uniques " + total);

	}
	
	public void DrawPoint(String loc, float radius, float percent)
	{
		grf.DrawPoint(loc, radius, percent);
	}
	
	public void GraphPointsI()
	{
		long total = 0;
		float max = 0;
		float min = 1000;
		grf.DrawKey();
		list = loadSegment(dataname);
		for (int i = list.size() - 1; i >= 0; i--) {
			GRMapDataItem item = list.get(i);
			if (item.uniques >= min_uniques) {
				if (item.internal_affinity > max) {
					max = item.internal_affinity;
				}
				if (item.internal_affinity < min) {
					min = item.internal_affinity;
				}
			}
		}
		float range = max - min;
		for (int i = list.size() - 1; i >= 0; i--) {
			GRMapDataItem item = list.get(i);
			if (item.uniques >= min_uniques) {
				float t = item.internal_affinity; 
				// normalize?
				// (item.internal_affinity - min) / range;
				grf.DrawPoint(item.data_name, item.radius, t);
				total += item.uniques;
			}
		}

		System.out.println("total uniques " + total);
	}
	
	public void GraphPointsT()
	{
		long total = 0;
		float max_tempo = 0;
		float min_tempo = 1000;
		grf.DrawKey();
		list = loadSegment(dataname);
		for (int i = list.size() - 1; i >= 0; i--) {
			GRMapDataItem item = list.get(i);
			if (item.uniques >= min_uniques) {
				if (item.tempo > max_tempo) {
					max_tempo = item.tempo;
				}
				if (item.tempo < min_tempo) {
					min_tempo = item.tempo;
				}
			}
		}
		float range = max_tempo - min_tempo;
		for (int i = list.size() - 1; i >= 0; i--) {
			GRMapDataItem item = list.get(i);
			if (item.uniques >= min_uniques) {
				float t = (item.tempo - min_tempo) / range;
				t = 1.0f - t;  // invert given 105% is slower than 95% for tempo
				grf.DrawPoint(item.data_name, item.radius, t);
				total += item.uniques;
			}
		}
		System.out.println("total uniques " + total);
	}
	
	public void GraphPointsE()
	{
		long total = 0;
		float max = 0;
		float min = 1000;
		grf.DrawKey();
		list = loadSegment(dataname);
		for (int i = list.size() - 1; i >= 0; i--) {
			GRMapDataItem item = list.get(i);
			if (item.uniques >= min_uniques) {
				if (item.expert > max) {
					max = item.expert;
				}
				if (item.expert < min) {
					min = item.expert;
				}
			}
		}
		float range = max - min;
		for (int i = list.size() - 1; i >= 0; i--) {
			GRMapDataItem item = list.get(i);
			if (item.uniques >= min_uniques) {
				float t = (item.expert - min) / range;
				grf.DrawPoint(item.data_name, item.radius, t);
				total += item.uniques;
			}
		}
		System.out.println("total uniques " + total);
	}
	
	public void GraphCities()
	{
		MPGeocoder.ReadCityList();
		ArrayList<MPCityItem> cities = MPGeocoder.GetCityList();
		for (MPCityItem city : cities) {
			grf.HighlightPoint(city.city_lat, city.city_long, 0.5);
		}		
	}
	
	private ArrayList<GRMapDataItem> loadSegment(String path)
	{
		ArrayList<GRMapDataItem> segment = new ArrayList<GRMapDataItem>();
		MPTxtReader txrdr = new MPTxtReader(path);
		String s;
		
		// skip first line
		s = txrdr.MPReadTxtLn();
		
		// second line has global parameters; second field is default radius
		String[] globals = txrdr.MPReadTxtLn().split("\t");
		default_radius = Float.valueOf(globals[1]);
		
		while ((s = txrdr.MPReadTxtLn()) != null && !s.isEmpty()) {
		    String[] tokens = s.split("\t");

		    // name is first
		    String[] name_values = tokens[NAME_INDEX].split(",");
		    String location = name_values[0] + "," + name_values[1];		    
		    float radius = Float.valueOf(tokens[RADIUS_INDEX]);
		    float inter_aff = Float.valueOf(tokens[INTER_INDEX]);
		    float intra_aff = Float.valueOf(tokens[INTRA_INDEX]);
		    int uniques = Integer.valueOf(tokens[UNIQUES_INDEX]);
		    float expert = Float.valueOf(tokens[EXPERT_INDEX]);
		    float tempo = Float.valueOf(tokens[TEMPO_INDEX]);
		    
		    GRMapDataItem item = new GRMapDataItem(location, radius, inter_aff, intra_aff, 
		    							uniques, expert, tempo);
		    segment.add(item);
		}
		txrdr.Close();
		return segment;
	}
	
	void TestCircles()
	{
		grf.TestCircles();
	}
}

