import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;

import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;
import java.net.URLEncoder;
import org.apache.commons.io.IOUtils;

public class GeoQuery {

	/*Geocode request URL. Here see we are passing "json" it means we will get the output in JSON format.
	* You can also pass "xml" instead of "json" for XML output.
	* For XML output URL will be "http://maps.googleapis.com/maps/api/geocode/xml"; 
	*/
	
	private static final String URL = "http://maps.googleapis.com/maps/api/geocode/json"; 
	
	/*
	* Here the fullAddress String is in format like "address,city,state,zipcode". Here address means "street number + route" .
	*
	*/
	public String getJSONByGoogle(String fullAddress) 
	{
	
		/*
		* Create an java.net.URL object by passing the request URL in constructor. 
		* Here you can see I am converting the fullAddress String in UTF-8 format. 
		* You will get Exception if you don't convert your address in UTF-8 format. Perhaps google loves UTF-8 format. :)
		* In parameter we also need to pass "sensor" parameter.
		* sensor (required parameter)  Indicates whether or not the geocoding request comes from a device with a location sensor. This value must be either true or false.
		*/
		// String api_key ="&key=" + "AIzaSyCGqg-U-IaJepcHDQbhgwLgkoJMAdB8BUU";
		URL url = null;
		try {
			url = new URL(URL + "?latlng=" + URLEncoder.encode(fullAddress, "UTF-8")+ "&sensor=true");
		} catch (UnsupportedEncodingException e) {
			System.err.println("unsupported URL exception");
			return null;
		} catch (MalformedURLException e) {
			System.err.println("mal-formed URL 	exception");
			return null;
		}
			
		// Open the Connection 
		URLConnection conn = null;
		try {
			conn = url.openConnection();
		} catch (IOException e) {
			return null;
		}
		
		// This is Simple a byte array output stream that we will use to keep the output data from google. 
		ByteArrayOutputStream output = new ByteArrayOutputStream(1024);
		
		// copying the output data from Google which will be either in JSON or XML depending on your request URL that in which format you have requested.
		try {
			IOUtils.copy(conn.getInputStream(), output);
		} catch (IOException e) {
			return null;
		}
		
		//close the byte array output stream now.
		try {
			output.close();
		} catch (IOException e) {
			return null;
		}
		
		return output.toString(); // This returned String is JSON string from which you can retrieve all key value pair and can save it in POJO.
	}
}
