import java.io.File;
import java.util.ArrayList;


/** 
 * MPAnalyzer
 */

public class MPAnalyzer {

	static MPConfig the_config = null;
	
	public static int GetPerfMax()
	{
		return GetConfig().performance_max;
	}
	
	public static boolean UseOriginalPerformance()
	{
		return GetConfig().use_original_performance;
	}
	
	public static boolean PruneNonGeoData()
	{
		return GetConfig().prune_non_geo_data;
	}
	
	public static boolean ExcludeUrban()
	{
		return GetConfig().exclude_urban;
	}
	
	public static boolean ExcludeRural()
	{
		return GetConfig().exclude_rural;
	}
	
	public static boolean PruneChords()
	{
		return GetConfig().prune_chords;
	}
	
	public static boolean UnpackChords()
	{
		return GetConfig().unpack_chords;
	}
	
	public static boolean ExcludeUS()
	{
		return GetConfig().exclude_us;
	}
	
	public static boolean ExcludeRepeats()
	{
		return GetConfig().exclude_repeats;
	}
	
	
	public static MPConfig GetConfig()
	{
		return the_config;
	}
	
	public static void main(String[] args) 
	{
		the_config = new MPConfig(args);
		the_config.Out();
			
		if (the_config.flush_cache) {
			MPGeocoder.FlushGeoCache();
		} else if (the_config.sn_analyze) {
			MPAnalyzeSNData();	
		} else if (the_config.gr_graph) {
			GraphMap();
		} else {
			MPInit();
			// MPMapAllLocations();
			// MPAnalyzeStoreMetaData();
			if (!the_config.use_ui) {
				if (the_config.compare_to_region) {
					MPGeoCoord2 coord = MPGeocoder.reverseGeocode(the_config.compare_to_region_str);
					MPAnalyzeSongs(the_config.regiontype, coord);
				} else {
					MPAnalyzeSongs(the_config.regiontype, null);
				}
			}
			MPGeocoder.WriteGeoCache();
		}
	}
	
	static public String GetPerfDirectory()
	{
		return GetConfig().perf_directory;
	}

	static public void MPMapAllLocations()
	{
		GRMapData data = new GRMapData();
		data.LoadGraph();

		MPFileDriver fd = new MPFileDriver(GetConfig().perfdriver);
		for (String the_path : fd.filepaths) {
			System.out.println(the_path);
			MPSong thesong = new MPSong(new File(the_path), GetPerfMax());
			thesong.LoadMeta();
			// thesong.LoadSong();
			System.out.println("rendering " + thesong.songIO.songFileName + " " + thesong.songList.getSize());
			thesong.Graph(data, 0.4f, 0.9f);
			thesong.Close();
		}
		MPGeocoder.WriteGeoCache();
		data.StoreGraph();
		
		System.out.println("done");
		
	}
	
	static public void MPAnalyzeStoreMetaData()
	{
		MPFileDriver fd = new MPFileDriver(GetConfig().perfdriver);
		for (String the_path : fd.filepaths) {
			System.out.println(the_path);
			MPSong thesong = new MPSong(new File(the_path), GetPerfMax());
			thesong.OutputSongMetaData();
		}
		System.out.println("done");
	}
	
	static public void MPAnalyzeSongs(int region_type, MPGeoCoord2 coord)
	{
		MPFileDriver fd = new MPFileDriver(GetConfig().perfdriver);

		RIRegionItemSet ri_tree = new RIRegionItemSet();
		ArrayList<MPSong> songs = new ArrayList<MPSong>();

		// Iterate over all directory items and load songs
		for (String the_path : fd.filepaths) {
			System.out.println(the_path);
			MPSong thesong = new MPSong(new File(the_path), GetPerfMax());
			thesong.LoadMeta();
			songs.add(thesong);
		}
		MPGeocoder.WriteGeoCache();

		ri_tree.AllocateRegions(region_type, songs);
		System.out.println("populating regions " + ri_tree.GetSize());
		
		// populate with a lot more performances depending on region type
		int num_populate_per_region = GetConfig().max_uniques * 20;
		if (region_type != MPConfig.ANALYZE_BY_GEO) {
			// num_populate_per_region *= 100;
			num_populate_per_region = GetConfig().max_uniques;
		}
		ri_tree.Populate(songs, num_populate_per_region);
		
		System.out.println("adjusting region sizes " + ri_tree.GetSize());
		if (region_type == MPConfig.ANALYZE_BY_GEO) {
			ri_tree.AdjustRegionSizes(GetConfig().min_uniques, GetConfig().max_uniques);
			System.out.println("populating adjusted regions " + ri_tree.GetSize());
			ri_tree.Populate(songs, GetConfig().max_uniques);

			// System.out.println("pruning " + ri_tree.GetSize());
			// ri_tree.PruneRedundantRegions();
			System.out.println("total regions " + ri_tree.GetSize());
		}
		

		if (coord != null) {
			ri_tree.CorrelateToRegion(coord, songs);
		} else {
			ri_tree.CorrelateByRegion(songs);
		}
					
		/*{
			MPOutputStream out = new MPOutputStream(GetConfig().the_file_target + "_p.xls");
			ri_tree.OutputTenRegions(out);
			out.MPClose();
		}*/

		MPOutputStream out = new MPOutputStream(GetConfig().the_file_target + ".xls");
		System.out.println("Report " + ri_tree.GetSize());
		ri_tree.Report(out, GetConfig().min_uniques);
		
		out.MPClose();
		System.out.println("done");
	}
	
	static public void MPInit()
	{
		// connect to our MySQL database
		boolean connected = MPSQL.connectToSQL();
		if ( !connected ) {
			System.out.println("Failed to connect to MySQL Database. Won't be able to download performances");
		} else {
			System.out.println("Connected To MySQL Database!");
		}
		
		// connect to the Amazon S3
		connected = MPAWS.connectToAWS();
		if ( !connected ) {
			System.out.println("Failed to connect to Amazon AWS. Won't be able to download performances");
		} else {
			System.out.println("Connected To Amazon AWS!"); 
		}
		
		// load the geocache first
		MPGeocoder.ReadGeoCache();		
		MPGeocoder.ReadCityList();
		
		if (GetConfig().use_ui) {
			MPPerfBrowser.initializeGUI();
			MPDialogs.initializeDialogs();
		}
	}
	
	public static void GraphMap()
	{
		GRMapData data = new GRMapData();
		data.LoadGraph();
		data.GraphPointsX();
		if (GetConfig().graph_cities) {
			data.GraphCities();
		}
		data.StoreGraph();
		
		System.out.println("done");
	}
	
	public static void GraphMapx()
	{
		GRMapData data = new GRMapData();
		data.LoadGraph();
		data.TestCircles();
		// data.GraphCities();
		data.StoreGraph();
		
		System.out.println("done");
	}
	

	public static void GraphMap3()
	{
		GRFrame frame = new GRFrame(51, 0, 102);
		if (frame.Load()) {
			frame.DrawPoint("80, 15", 1.0, (float).35);
			frame.DrawPoint("40, 0", 1.0, (float).65);
			frame.DrawPoint("0, 0", 1.0, (float).85);
						
			// tokyo
			frame.DrawPoint("35.68994, 139.69170", 1.0, .80f);
			
			// aukland
			frame.DrawPoint("-36.848395, 174.763314", 1.0, .80f);
			
			// san diego
			frame.DrawPoint("32.715786, -117.158340", 1.0, .80f);			
			
			int j = 0;
			float percent = .8f;
			for (int i = -106; i <= -87; i++) {
				frame.DrawPoint("20," + i, 1.0, percent );
				percent -= .02f;
				++j;
			}
			
			System.out.println(j);

			frame.Store("2.png");
		}
		
		System.out.println("done");
	}
	static public void MPShutdown()
	{
	}

}
