
public class MPConfig {
	protected int performance_max = 200;
	protected int min_uniques = 22;
	protected int max_uniques = 44;

	protected boolean flush_cache = false;
	protected int GEO_QUERY_MAX = 1;
	protected float GEO_SCALE_APPROX = (float)2000.0;

	protected static final int ANALYZE_BY_COUNTRY = 1;
	protected static final int ANALYZE_BY_GEO = 2;
	protected static final int ANALYZE_BY_STATE = 3;
	protected static final int ANALYZE_BY_CITY = 4;
	protected static final int ANALYZE_BY_CONTINENT = 5;
	protected static final int ANALYZE_BY_LATITUDE = 6;
	protected int regiontype = ANALYZE_BY_GEO;

	protected float geo_region_size = 50.0f; // in sq km
	protected float default_region_radius = 2.0f; // in degrees long/lat
	protected float geo_region_scale = 3.0f; // applied to radius 
	protected int ANALYZE_SEGMENT_MAX = 100000;

	protected boolean compare_to_region = false;
	protected String compare_to_region_str = "15.0,100.0";

	protected boolean gr_graph = false;
	protected boolean graph_cities = false;
	protected String the_file_target = "shows2";
	protected String map_file_name = "../image/gv3.png";
	protected boolean sn_analyze = false;
	protected String perfdriver = "perffiles.txt";
	protected String perf_directory = "../perfs/xmas/";

	protected boolean use_ui = false;
	protected boolean use_original_performance = false;
	protected boolean exclude_us = false;
	protected boolean exclude_urban = false;
	protected boolean exclude_rural = false;
	protected boolean prune_chords = false;
	protected boolean unpack_chords = false;
	protected boolean exclude_repeats = false;
	protected boolean prune_non_geo_data = true;

	public MPConfig(String[] args)
	{
		String config_path = "mpconfig.json";
		loadJsonConfig(config_path);
		parseArgs(args);
	}
	
	public void parseArgs(String[] args)
	{
		if (args != null && args.length > 0) {
			gr_graph = true;
			the_file_target = args[0];
		}
	}

	private void loadJsonConfig(String path)
	{
		MPJson json = new MPJson();
		if (!json.MPJsonOpenFile(path)) {
			return;
		}

		performance_max = json.MPJsonGetInt("performance_max");
		min_uniques = json.MPJsonGetInt("min_uniques");
		max_uniques = json.MPJsonGetInt("max_uniques");
		exclude_repeats = json.MPJsonGetBoolean("exclude_repeats");
		unpack_chords = json.MPJsonGetBoolean("unpack_chords"); 
		prune_chords = json.MPJsonGetBoolean("prune_chords"); 

		flush_cache = json.MPJsonGetBoolean("flush_cache");
		GEO_QUERY_MAX = json.MPJsonGetInt("GEO_QUERY_MAX");
		GEO_SCALE_APPROX = (float)json.MPJsonGetDouble("GEO_SCALE_APPROX");

		regiontype = json.MPJsonGetInt("regiontype");
		geo_region_size = (float)json.MPJsonGetDouble("geo_region_size");
		default_region_radius = (float)json.MPJsonGetDouble("default_region_radius");
		geo_region_scale = (float)json.MPJsonGetDouble("geo_region_scale");

		ANALYZE_SEGMENT_MAX = json.MPJsonGetInt("ANALYZE_SEGMENT_MAX");

		compare_to_region = json.MPJsonGetBoolean("compare_to_region");
		compare_to_region_str = json.MPJsonGetString("compare_to_region_str");

		gr_graph = json.MPJsonGetBoolean("gr_graph");
		graph_cities = json.MPJsonGetBoolean("graph_cities");
		use_ui = json.MPJsonGetBoolean("use_ui");
		use_original_performance = json.MPJsonGetBoolean("use_original_performance");

		the_file_target = json.MPJsonGetString("file_target");
		map_file_name = json.MPJsonGetString("map_file_name");
	
		sn_analyze = json.MPJsonGetBoolean("sn_analyze");
		perfdriver = json.MPJsonGetString("perfdriver");
		perf_directory = json.MPJsonGetString("perf_directory");

		json.MPJsonCloseFile();
		
		if (prune_chords && !use_original_performance) {
			System.err.println("can not prune chords without original performance file");
		}
	}
	
	public boolean useFullGeoCoordinates()
	{
		return regiontype != ANALYZE_BY_GEO;
	}
	
	public void Out()
	{
		System.out.println("config perf-max " + performance_max + " " +
				"min " + min_uniques + " " +
				"max " + max_uniques + " " +
				"ex-repeats " + exclude_repeats + " " +
				"region-size " + geo_region_size + " " +
				"region-radius " + default_region_radius + " " +
				"region-scale " + geo_region_scale + " " +
				"compare-to-region " + compare_to_region + " " +
				the_file_target );
		
	}

}
