import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.*;

public class MPGeocoder {
	static int over_query_limit_error_cnt = 0;
	static int QUERY_MAX;
	static final String cachefilename = "geocache" + File.separator + "mpgeocache.data";
	static final String citylistname = "geocache" + File.separator + "cities.txt";

	static public class MPGeoCacheHeader implements Serializable
	{
		public static final long serialVersionUID = 1;
		public long headercount = 0;
		
		public MPGeoCacheHeader()
		{
		}
	}
		
	static ArrayList<MPGeoCoord2> cache = null;
	static ArrayList<MPCityItem> citylist = null;
	
	static boolean cachedirty = false;
	
	private static short writeCache(ObjectOutputStream oos)
	{
		short result = 0;
		try {
			
			//MPGeoCacheHeader header = new MPGeoCacheHeader();
			// header.headercount = cache.size();
			// oos.writeObject(header);		
			oos.writeObject(cache);
			
			/*
	        Iterator<MPGeoCoord2> itr = cache.iterator();

	        while (itr.hasNext()) {
	            Object element = itr.next();
	        	MPGeoCoord2 value = (MPGeoCoord2)element;
				oos.writeObject(value);
	        }
	        */
	        
	        System.out.println("geo cache write " + cache.size());

		} catch (IOException e) {
			result = -1;
			System.out.println("writeCache: IO exception " + e);			
		}
		
		return result;
	}
	
	@SuppressWarnings("unchecked")
	private static short readCache(ObjectInputStream ois)
	{
		short result = 0;
		
		try {
			cache = (ArrayList<MPGeoCoord2>)ois.readObject();
			/*
			MPGeoCacheHeader header = (MPGeoCacheHeader)ois.readObject();
			
			cache = new ArrayList<MPGeoCoord2>((int)header.headercount);
			for (int i = 0; i < header.headercount; i++) {
				MPGeoCoord2 item = (MPGeoCoord2)ois.readObject();
				cache.add(item);
			}
			*/
	        System.out.println("geo cache read " + cache.size());

		} catch (IOException e) {
			System.out.println("IO exception " + e);
			result = -1;
			
		} catch (ClassNotFoundException e) {
			System.out.println("Class not found exception " + e);
			result = -1;
		}
		
		return result;
	}
	
	public static short ReadGeoCache()
	{
		QUERY_MAX = MPAnalyzer.GetConfig().GEO_QUERY_MAX;
		
		// only read the cache if we need it
		if (!MPAnalyzer.GetConfig().useFullGeoCoordinates()) {
			return 0;
		}

		short result = -1;
		
		File cachefile = new File(cachefilename);
		
		try {
			FileInputStream fis = new FileInputStream(cachefile);
			ObjectInputStream ois = new ObjectInputStream(fis);
			result = readCache(ois);
			ois.close();
		} catch (FileNotFoundException e) {
			System.err.println("File Not Found exception " + e);
		} catch (IOException e) {
			System.err.println("IO exception " + e);
		}
		
		if (result != 0) {
			cache = new ArrayList<MPGeoCoord2>(0);
		}
		
		return result;
	}
	
	public static short FlushGeoCache()
	{
		short result = -1;
		if (ReadGeoCache() != 0) {
			return result;
		}
				
		// reset city values (in case definition changed)
		ReadCityList();
		for (MPGeoCoord2 coord : cache) {
			MPCityItem ci = GetCity(coord);
			if (ci != null) {
				coord.is_city = true;
				coord.commonLocalityName = ci.GetName();
			}
		}
		
		Collections.sort(cache);
		cachedirty = true;
		result = WriteGeoCache();
		
		return result;
	}

	public static short FlushGeoCache3()
	{
		short result = -1;
		if (ReadGeoCache() == 0) {
			int cnt = 0;
			for (int i = 0; i < cache.size(); i++) {
				MPGeoCoord2 c = cache.get(i);
				
				// try to repair cache if country name is null
				if (c != null && !exceededQueryLimit()) {
					if (c.administrativeAreaName != null && c.administrativeAreaName.contentEquals("")) {
					
						if (executeReverseGeocode(c, String.valueOf(c.n_lat), String.valueOf(c.n_long))) {
							if (c.administrativeAreaName != null && c.administrativeAreaName.contentEquals("")) {
								cache.set(i, c);
								++cnt;
								System.out.println(cnt);
								if (cnt % 20 == 0) {
									cachedirty = true;
									result = WriteGeoCache();
								}
							} else {
								System.out.println("name still null");
							}
						} else {
							System.out.println("fail: " + c.n_lat + "," + c.n_long);
						}
					}
				}
			}
			Collections.sort(cache);
			cachedirty = true;
			result = WriteGeoCache();
		}
		if (result != 0) {
			System.err.println("flush cache failed");
		}
		return result;
	}
	
	public static short WriteGeoCache()
	{
		if (!cachedirty) {
			return 0;
		}
		
		if (!MPAnalyzer.GetConfig().useFullGeoCoordinates()) {
			System.err.println("geo error:  cache dirty without full coordinates");
		}
		
		short result = -1;
		// load meta data from file
		try {
			File cachefile = new File(cachefilename);
			FileOutputStream fos = new FileOutputStream(cachefile);
			ObjectOutputStream oos = new ObjectOutputStream(fos);
			result = writeCache(oos);
			oos.close();
		} catch (FileNotFoundException e) {
			System.out.println("File Not Found exception " + e);
		} catch (IOException e) {
			System.out.println("IO exception " + e);
		}
		
		if (result == 0) {
			cachedirty = false;
		}
		
		return result;
	}

	private static void geoCacheAdd(int index, MPGeoCoord2 coord)
	{
		cachedirty = true;
		cache.add(index, coord);		
	}
	
	private static boolean exceededQueryLimit()
	{
		return over_query_limit_error_cnt > QUERY_MAX;
	}
	
	private static int cache_hit_count = 0;
	private static int cache_miss_count = 0;
	
	public static MPGeoCoord2 reverseGeocode(String coord)
	{
		String tokens[] = coord.split(",");
		return reverseGeocode(tokens[0], tokens[1]);
	}

	public static MPGeoCoord2 reverseGeocode(String latitude, String longitude)
	{
		MPGeoCoord2 geoCoord = new MPGeoCoord2(latitude, longitude);
		if (!MPAnalyzer.GetConfig().useFullGeoCoordinates() || geoCoord.isGeoCodeNull()) {
			return geoCoord;
		}
				
		int index = Collections.binarySearch(cache, geoCoord);
		if (index >= 0) {
			++cache_hit_count;
			// System.out.println("cache hit " + index);
			MPGeoCoord2 cachehit = cache.get(index);
			return cachehit;
		}
		
		++cache_miss_count;
		// System.out.println("cache miss " + index + " " + latitude + "," + longitude);
		if (cache_miss_count % 100 == 0) {
			System.out.println("miss/hit " + cache_miss_count + " " + cache_hit_count);
		}

		if (!exceededQueryLimit()) {
			boolean result = executeReverseGeocode(geoCoord, latitude, longitude);
			if (result) {
				geoCacheAdd(-index - 1, geoCoord);
			
			// if this last query pushed us over the top, then start approximating
			} else if (exceededQueryLimit()) {
				System.out.println("\nhit query limit\n");
				WriteGeoCache();
			}
		}
		
		return geoCoord;
	}
			
	public static boolean executeReverseGeocode(MPGeoCoord2 geoCoord, String latitude, String longitude)
	{
		// technically should never fire
		if (exceededQueryLimit()) {
			return false;
		}
		
		GeoQuery gci = new GeoQuery();
		String queryresult = gci.getJSONByGoogle(latitude + "," + longitude);
		if (queryresult == null) {
			System.out.println("reverse geocode fail:  null");
			return false;
		}
		
		MPJson mpj = new MPJson();
		mpj.MPJsonOpenString(queryresult);
		
		String status = mpj.MPJsonGetString("status");
		if (status.compareToIgnoreCase("OK") != 0) {
			System.out.println("reverse geocode fail " + status);
			if (status.compareToIgnoreCase("OVER_QUERY_LIMIT") == 0) {
				++over_query_limit_error_cnt;
				// try to appease google by slowing down request rate
				//try {
				//    Thread.sleep(200);
				//} catch(InterruptedException ex) {
				//    Thread.currentThread().interrupt();
				//}			
			}
			return false;
		}
		
		ArrayList<String> results = mpj.MPJsonGetStringArray("results");
		if (results == null) {
			System.out.println("reverse geocode fail");
			return false;
		}
		
		// the first string of results should be the address and its components
		return parseJsonResults(geoCoord, results.get(0));
	}
	
	/**
	 * we're passed a string in format of google reverse geocode v3.0 
	 * https://developers.google.com/maps/documentation/geocoding/#ReverseGeocoding
	 * 
	 * parse the string and populate the appropriate fields in geoCoord
	 * 
	 * we first find "address_components" field which includes street, locality, etc.
	 * we then look for "formatted_address" field which contains the full formatted address.
	 * 
	 */
	
	public static boolean parseJsonResults(MPGeoCoord2 geoCoord, String results)
	{
		MPJson mpj = new MPJson();
		mpj.MPJsonOpenString(results);
		
		geoCoord.address = mpj.MPJsonGetString("formatted_address");
		
		ArrayList<String> addr_components = mpj.MPJsonGetStringArray("address_components");
		String street_number = "";
		for (int i = 0; i < addr_components.size(); i++) {
			MPJson mpjsub = new MPJson();
			mpjsub.MPJsonOpenString(addr_components.get(i));
			String long_name = mpjsub.MPJsonGetString("long_name");
			String short_name = mpjsub.MPJsonGetString("short_name");
			String name = long_name;
			if (name == null) {
				name = short_name;
			}
			ArrayList<String> component_types = mpjsub.MPJsonGetStringArray("types");
			if (component_types != null) {
				for (int j = 0; j < component_types.size(); j++) {
					String s = component_types.get(j);
					if (s.equalsIgnoreCase("street_number")) {
						street_number = name;
					} else if (s.equalsIgnoreCase("locality")) {
						geoCoord.localityName = name;
					} else if (s.equalsIgnoreCase("route")) {
						geoCoord.thoroughfareName = street_number + " " + name;
					} else if (s.equalsIgnoreCase("administrative_area_level_1")) {
						geoCoord.administrativeAreaName = name;
					} else if (s.equalsIgnoreCase("country")) {
						geoCoord.countryName = long_name;
						geoCoord.countryNameCode = short_name;
					} else if (s.equalsIgnoreCase("postal_code")) {
						geoCoord.postalCodeNumber = name;
					}
				}
			}
		}
		
		// finally, cache whether this coord is a city or not
		// note we must flush the cash if this definition changes
		MPCityItem ci = GetCity(geoCoord);
		if (ci != null) {
			geoCoord.is_city = true;
			geoCoord.commonLocalityName = ci.GetName();
		}
			
		return geoCoord.address != null && !geoCoord.address.isEmpty();
	}
	
	
	public static ArrayList<MPCityItem> ReadCityList()
	{		
		MPTxtReader txt = new MPTxtReader(citylistname);
		citylist = new ArrayList<MPCityItem>();
		
		String s;
		while ((s = txt.MPReadTxtLn()) != null) {
			MPCityItem ci = new MPCityItem(s);
			citylist.add(ci);
		}
		
		return citylist;
	}
	
	public static boolean IsCity(MPGeoCoord2 coord) 
	{
		return GetCity(coord) != null;
	}
	
	public static MPCityItem GetCity(MPGeoCoord2 coord) 
	{
		MPCityItem c1 = new MPCityItem(coord.n_lat, coord.n_long);
		for (MPCityItem c2 : citylist) {
			if (c1.RadiusMatch(c2)) {
				return c2;
			}
		}
		
		return null;
	}
	
	public static ArrayList<MPCityItem> GetCityList()
	{
		return citylist;
	}
}

