import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.ArrayList;

public class MPMidiData {
	
	/** An array of notes hit during performance, ordered chronologically */
	protected ArrayList<MPNoteEvent> noteEvents;

	/** The tempo of this performance in BPM */
	public long BPM = -1;
	
	protected boolean valid_midi = false;

	static public class MPMidiDataHeader implements Serializable {
		public static final long serialVersionUID = 1;
		public long headercount;
		public long checksum;
		public long BPM;

		public MPMidiDataHeader()
		{
		}
	}

	public MPMidiData(File midiFile, float bpmRate, boolean cache)
	{
		if (cache) {
			File cachefile = new File(midiFile.getPath().substring(0, midiFile.getPath().length() - 5) + ".cache");
	
			if (ReadMidiCache(cachefile) == 0) {
				valid_midi = true;
				return;
			}
			
			noteEvents = new ArrayList<MPNoteEvent>(0);
			if (MPMIDI.parseMIDI(this, bpmRate, midiFile) == 0) {
				WriteMidiCache(cachefile);
				valid_midi = true;
			}
		} else {
			noteEvents = new ArrayList<MPNoteEvent>(0);
			if (MPMIDI.parseMIDI(this, bpmRate, midiFile) == 0) {
				valid_midi = true;
			}
		}
	}
	
	protected boolean isValid()
	{
		return valid_midi;
	}
	
	protected void Clear()
	{
		noteEvents.clear();
	}
	
	private short writeCache(ObjectOutputStream oos)
	{
		short result = 0;
		try {
			
			MPMidiDataHeader header = new MPMidiDataHeader();
			header.headercount = noteEvents.size();
			header.BPM = BPM;
			header.checksum = 1;
			oos.writeObject(header);	
			oos.writeObject(noteEvents);

		} catch (IOException e) {
			result = -1;
			System.err.println("writeCache: IO exception " + e);			
		}
		
		return result;
	}
	
	@SuppressWarnings("unchecked")
	private short readCache(ObjectInputStream ois)
	{
		short result = 0;
		
		try {
			MPMidiDataHeader header = (MPMidiDataHeader)ois.readObject();
			BPM = header.BPM;
			if (header.checksum == 1) {
				noteEvents = (ArrayList<MPNoteEvent>)ois.readObject();
			} else {
				result = -1;
			}

		} catch (IOException e) {
			System.err.println("IO exception " + e);
			result = -1;
			
		} catch (ClassNotFoundException e) {
			System.err.println("Class not found exception " + e);
			result = -1;
		}
		
		return result;
	}

	public short ReadMidiCache(File cachefile)
	{
		short result = -1;
		
		try {
			FileInputStream fis = new FileInputStream(cachefile);
			ObjectInputStream ois = new ObjectInputStream(fis);
			result = readCache(ois);
			ois.close();
		} catch (FileNotFoundException e) {
		} catch (IOException e) {
		}
		
		return result;
	}
	
	public short WriteMidiCache(File cachefile)
	{
		short result = -1;
		// load meta data from file
		try {
			FileOutputStream fis = new FileOutputStream(cachefile);
			ObjectOutputStream oos = new ObjectOutputStream(fis);
			result = writeCache(oos);
			oos.close();
		} catch (FileNotFoundException e) {
			System.out.println("File Not Found exception " + e);
		} catch (IOException e) {
			System.out.println("IO exception " + e);
		}
		
		return result;
	}

}

