import java.io.File;
import java.util.*;

public class MPOriginalPerformance {
	
	protected MPMidiData midi;
	
	/** BPM Rate -- defined in original json performance file **/
	protected float bpmRate = (float)1.0;
	
	protected ArrayList<Long> noteDur;
	
	// reference to parent song input/output class
	private MPSongIO songIO;
		
	/** The file associated with this performance */
	public File performanceFile;
	
	protected float noteAverageDuration;
	protected float noteStdDeviation;
	
	protected float perfSlope;
	protected ArrayList<Float> noteSlope;
	
	protected long sortNoteValue;
	protected boolean is_pruned;

	protected void Clear()
	{
		if (noteDur != null) {
			noteDur.clear();
		}
		if (noteSlope != null) {
			noteSlope.clear();
		}
		if (midi != null) {
			midi.Clear();
		}
		is_pruned = true;
	}
	
	protected void loadMidi(float rate, boolean cache)
	{
		// instantiate the note event array
		midi = new MPMidiData(performanceFile, rate, cache);
	}
	
	public MPOriginalPerformance(MPSongIO sIO, File midiFile)
	{
		songIO = sIO;
		performanceFile = midiFile;
	}
	
	public MPOriginalPerformance(MPSongIO sio, File midiFile, float rate)
	{
		songIO = sio;
		performanceFile = midiFile;
		bpmRate = rate;
	}
	
	public boolean LoadPerformance()
	{
		boolean cache_the_performance = true;
	
		// instantiate the note event array
		loadMidi(bpmRate, cache_the_performance);
		if (midi.valid_midi) {
			analyzePerfData();
			return true;
		}
		return false;
	}
	
	public String GetPerformanceSongID()
	{
		return songIO.songFileName;
	}
		
	public long getNoteDur(int i)
	{
		if (i >= noteDur.size()) {
			return 0;
		}
		return noteDur.get(i);
	}
	
	public void SwapNotes(int k, int l)
	{
		long temp = noteDur.get(k);
		noteDur.set(k, noteDur.get(l));
		noteDur.set(l, temp);
		MPNoteEvent t2 = midi.noteEvents.get(k);
		midi.noteEvents.set(k, midi.noteEvents.get(l));
		midi.noteEvents.set(l, t2);
	}
	
	public float getNoteSlope(int i)
	{
		return noteSlope.get(i);
	}
	
	public int getNumNotes()
		{
			return noteDur.size();
		}
	
	public void setSlope(Float s)
	{
		perfSlope = s;
	}
	
	public Float getSlope()
	{
		return perfSlope;
	}
	
	protected void pruneNonChords()
	{
		int[] indexes = songIO.getChordIndex();
		if (indexes == null) {
			return;
		}
		
		ArrayList<MPNoteEvent> chordEvents = new ArrayList<MPNoteEvent>();
		for (int i = 0; indexes[i] != 0; i++) {
			chordEvents.add(i, midi.noteEvents.get(indexes[i]));
		}
		
		midi.noteEvents = chordEvents;
	}
	
	// insufficient to just find the string.  must find the exact size string.
	
	private int findConsecutive(int i, long value) 
	{
		int j = i;
		while (i < noteDur.size() && noteDur.get(i) == value) {
			++i;
		}
		return i - j;
	}
	
	public int CalcDifficulty(int consecutive, long value) 
	{
		int count = 0;
		int i = 0;
		while (i < noteDur.size()) {
			int c = findConsecutive(i, value);
			if (c == consecutive) {
				++count;
			}
			// skip all strings regardless of size
			i += c + 1;
		}
		
		return count;
	}
	
	public int CalcDifficulty(int consec1, int consec2, long value) 
	{
		int count = 0;
		for (int c = consec1; c <= consec2; c++) {
			count += CalcDifficulty(c, value);
		}
		return count;
	}
	
	/*
	 * custom sorting value
	 */
	
	protected void SetSortValue(long value)
	{
		sortNoteValue = value;
	}
	
	protected long GetSortValue()
	{
		return sortNoteValue;
	}

	/*
	// Chords!!!
	
	public int calcMistakes(MPOriginalPerformance p)
	{
		int mistakes = 0;
		String ss1 = "";
		String ss2 = "";
		int count = Math.min(noteEvents.size(), p.noteEvents.size());
		count = Math.min(count, 41);
		for (int i = 0; i < count; i++) {
			MPNoteEvent e1 = noteEvents.get(i);
			MPNoteEvent e2 = p.noteEvents.get(i);
			String s1 = e1.getPitch();
			String s2 = e2.getPitch();
			ss1 += s1 + " ";
			ss2 += s2 + " ";
			if (!s1.contentEquals(s2)) {
				++mistakes;
			}
		}
		
		System.out.println(mistakes);
		System.out.println(ss1);
		System.out.println(ss2);
		return mistakes;
	}
	*/
	
	/*
	 * CalcSlope:  the delta in the average note duration for this performance
	 * vs aveNoteDur -- that of the entire song.
	 * 
	 */
	
	public float CalcSlope(ArrayList<Float> aveNoteDur)
	{
		perfSlope = 0;
		noteSlope = new ArrayList<Float>();
		float delta = 0;

		int min = Math.min(getNumNotes(), aveNoteDur.size());
		for (int j = 0; j < min; j++) {
			float dur = (float)getNoteDur(j);
			
			// delta > 0 => faster than average
			delta = aveNoteDur.get(j) - dur;
			perfSlope += delta;
			noteSlope.add(perfSlope);
		}
		
		return perfSlope;
	}
	
    public int SongLength()
	{
		return midi.noteEvents.size();
	}
    	
    public void calcPerfStats()
    {
		int count = 0;
		float sum = 0;
		float stddev = 0;
		
		for (Long dur : noteDur) {
			sum += dur;
			stddev += dur * dur;
			count++;
		}
				
		// calculate the average & std dev
		float average = sum / (float)count;
		noteAverageDuration = average;
		average = average * average;
		float n = ((float)count) / (count - 1);
		average = n * average;
		stddev = stddev / (float)(count - 1);
		stddev = (float)Math.sqrt(stddev - average);
		noteStdDeviation = stddev;
    }
    
	/**
	 * Method to analyze the average time between notes
	 */
	public void analyzeTimeBetweenNotes()
	{
		int count = 0;
		float sum = 0;
		float stddev = 0;
		
		if (noteDur != null) {
			System.out.println("note dur non-null " + noteDur.size());
		}
		
		noteDur = new ArrayList<Long>();

		// iterate through all note events
		if (midi.noteEvents == null || midi.noteEvents.isEmpty()) {
			System.out.println("error: no note events \n" + performanceFile.getName());
			return;
		}
		
		long prevtime = midi.noteEvents.get(0).getTimeMS();
		for (int i = 1; i < midi.noteEvents.size(); i++) {
		// for (int i = 1; i < 41; i++) {
			MPNoteEvent event = midi.noteEvents.get(i);
			long curtime = event.getTimeMS();
			long dur = (curtime - prevtime);
			sum += dur;
			stddev += dur * dur;
			noteDur.add(dur);

			prevtime = curtime;
			count++;
		}
		
		// calculate the average & std dev
		float average = sum / (float)count;
		noteAverageDuration = average;
		average = average * average;
		float n = ((float)count) / (count - 1);
		average = n * average;
		stddev = stddev / (float)(count - 1);
		stddev = (float)Math.sqrt(stddev - average);
		noteStdDeviation = stddev;
//		System.out.println("Average Time Between Notes: " + average + "ms");
	}
	
	protected void analyzePerfData()
	{
		if (midi.BPM <= 0) {
			System.out.println("MPPerformance:AnalyzePerfData:  Error:  no tempo " + midi.BPM);
			return;
		}
		analyzeTimeBetweenNotes();
	}
	
	
	/*
	 * 
	 * Calc the PearsonCorrelation between two performances.  Using the following formula:
	 * 
	 * http://en.wikipedia.org/wiki/Spearman%27s_rank_correlation_coefficient
	 * 
	 * result should be between -1.0 and 1.0
	 * 
	 */
	public float SpearmanCorrelation(MPPerformance perf)
	{
		MPSpearmanArray sArray = new MPSpearmanArray(noteDur, perf.noteDur);
		return sArray.CalcSpearmanRank();
	}
	
	/*
	 * 
	 * 
	 */
	public float PearsonCorrelation(MPPerformance perf)
	{
		return SpearmanCorrelation(perf);
		/*
		int numNotes = getNumNotes();
		if (numNotes != perf.getNumNotes()) {
			return (float)0.0;
		}
		double mean = 0;
		for (int i = 0; i < numNotes; i++) {
			float x = noteDur.get(i) - noteAverageDuration;
			float y = perf.noteDur.get(i) - perf.noteAverageDuration;
			mean += x * y;
		}
		double r = Math.sqrt(mean / (numNotes * numNotes));	
		
		// float r2 = pearsonCorrelation(perf);
		// System.out.println(r + "\t" + r2);
		return (float)r;
		*/
	}
	
	/*
	 * 
	 * Calc the PearsonCorrelation between two performances.  Using the following formula:
	 * 
	 * http://en.wikipedia.org/wiki/Pearson_correlation
	 * 
	 * result should be between -1.0 and 1.0
	 * 
	 */
	public float PearsonCorrelationX(MPPerformance perf)
	{
		int numNotes = getNumNotes();
		if (numNotes != perf.getNumNotes()) {
			return (float)-1.0;
		}
		double mean = 0;
		double x_stddev = 0;
		double y_stddev = 0;
		//boolean same = true;
		for (int i = 0; i < numNotes; i++) {
			float x = noteDur.get(i) - noteAverageDuration;
			float y = perf.noteDur.get(i) - perf.noteAverageDuration;
			//if (same) {
			//	same = noteDur.get(i) == perf.noteDur.get(i);
			//}
			mean += x * y;
			x_stddev += x * x;
			y_stddev += y * y;
		}
		double stddev = Math.sqrt(x_stddev) * Math.sqrt(y_stddev);
		double r = mean / stddev;	
		
		// float r2 = pearsonCorrelation(perf);
		// System.out.println(r + "\t" + r2);
		//if (same) {
		//	System.out.println("identical performances " + r + " " + ++identical_count + " " + 
		//			performanceFile.getPath() + " " + perf.performanceFile.getPath());
		//}
		
		return (float)r;
	}
	
	/***
	 * 
	 * computationally cheaper version but not as accurate.
	 */
	public float PearsonCorrelation2(MPPerformance perf)
	{
		int numNotes = getNumNotes();
		if (numNotes != perf.getNumNotes()) {
			System.err.println("how did it get through?");
		}
		double mean = 0;
		for (int i = 0; i < numNotes; i++) {
			double x = noteDur.get(i) - noteAverageDuration;
			x = x / noteStdDeviation;
			
			double y = perf.noteDur.get(i) - perf.noteAverageDuration;
			y = y / perf.noteStdDeviation;

			mean += x * y;
		}
		
		double r = (1 / (float)(numNotes - 1)) * mean;
		return (float)r; 
	}

	/**
	 * Helper method to print out all the notes in this performance
	 */
	public void printAllNotes()
	{
		System.out.println("Print all notes");
		int cnt = 0;
		for (MPNoteEvent event : midi.noteEvents) {
			// print out the pitch and duration for each note
			// long time = event.getTimeMS(BPM);
			String pitch = event.getPitch();
			System.out.println(cnt + " " + pitch + " " + getNoteDur(cnt));
			++cnt;
		}
	}

}
