
/***
 * 
 * @author jeffreysmith
 *
 * a cluster of performances correlation values, indexed by perf (x,y)
 */
public class MPPerfCluster {

	protected MPPerfClusterCache clusterCache;
	
	RIRegionData clusterData;
		
	public MPPerfCluster(MPSongList list)
	{
		clusterCache = new MPPerfClusterCache(list.songs.size());
		clusterData = new RIRegionData();
		
		// use a cheap index into songs to define unique id for each song
		for (int i = 0; i < list.songs.size(); i++) {
			list.songs.get(i).setUniqueID2(i);
		}
	}
	
	public RIRegionData GetData()
	{
		return clusterData;
	}
		
	public void Clear()
	{
		if (clusterCache != null) {
			clusterCache.Clear();
			clusterCache = null;
		}
	}
		
	private float addItem(MPPerformance a, MPPerformance b)
	{
		// float r = a.DotProduct(b);
		float r = a.PearsonCorrelation(b);
		// float r = a.SpearmanCorrelation(b); 
		// float s = 0; // a.SpearmanCorrelation(b);
		clusterCache.add(a, b, r);
		return r;
	}
	
	public float Get(MPPerformance a, MPPerformance b)
	{
		float r;
		if (clusterCache.exists(a, b)) {
			r = clusterCache.get(a,b);
		} else {
			r = addItem(a, b);
		}
		return r;
	}

	public float GetPearson(MPPerformance a, MPPerformance b)
	{
		return Get(a,b);
	}
	
	public float GetSpearman(MPPerformance a, MPPerformance b)
	{
		return Get(a,b);
	}
	
	/***
	 * @param list
	 * @return
	 */
	
	public boolean CalcAvePearson(MPSongList list1, MPSongList list2)
	{
		long count = 0;
		double average = 0;
		for (MPPerformance a : list1.songs) {
			for (MPPerformance b : list2.songs) {
				if (!a.equals(b)) {
					average += Get(a,b);
					count++;
				}
			}
		}
		
		if (count > 0) {
			clusterData.pearson = average / (double)count;
			clusterData.compares = count;
			clusterData.size = list1.getSize() + list2.getSize();
			return true;
		}
		
		return false;
	}
	
	public boolean CalcAvePearson(MPSongList list)
	{
		long count = 0;
		double average = 0;
		for (int i = 0; i < list.getSize(); i++) {
			MPPerformance a = list.Get(i);
			for (int j = i + 1; j < list.getSize(); j++) {
				MPPerformance b = list.Get(j);
				average += Get(a,b);
				count++;
			}
		}
		
		if (count > 0) {
			clusterData.pearson = average / (double)count;
			clusterData.compares = count;
			clusterData.size = list.getSize();
			return true;
		}
		
		return false;
	}
	
	public float CalcAveSpearman(MPSongList list) 
	{
		System.out.println("no spearman numbers");
		return 0f;
	}
		
}

