import java.util.ArrayList;
import java.util.Collections;

public class MPRegionDataList {
	
	public class Quads {
		protected int the_quads[] = null;
	
		public Quads(int quad)
		{
			the_quads = new int[5];
			for (int i = 0; i < 5; i++) {
				the_quads[i] = 0;
			}
			if (quad > 0) {
				the_quads[quad]++;
			}
		}
		
		public Quads(int quads[])
		{
			the_quads = quads;
		}
		
		public void Add(Quads q)
		{
			for (int i = 1; i < 5; i++) {
				the_quads[i] += q.the_quads[i];
			}
		}
		
		public void Report(MPOutputStream out)
		{	
			int sum = 0;
			for (int i = 1; i < 5; i++) {
				sum += the_quads[i];
			}
			for (int i = 1; i < 5; i++) {
				out.MPWrite(the_quads[i] / (float)sum + "\t");
			}
		}
	}
	
	public class MPRegionDataItem implements Comparable<MPRegionDataItem>  {
		protected String region = null;
		protected Quads quads = null;
		protected int experts = 0;
		protected int amateurs = 0;
		protected double notedur = 0;
		protected int notecount = 0;
		
		public MPRegionDataItem(String reg)
		{
			region = reg;
		}

		public MPRegionDataItem(String reg, int quad, int expert, int amateur)
		{
			region = reg;
			quads = new Quads(quad);
			experts = expert;
			amateurs = amateur;
		}

		public void Add(MPRegionDataItem item)
		{
			quads.Add(item.quads);
			experts += item.experts;
			amateurs += item.amateurs;
			notedur += item.notedur;
			notecount += item.notecount;
		}
		
		public Quads getQuads()
		{
			return quads;
		}
		
		public int getExperts()
		{
			return experts;
		}
		
		public int getTotal()
		{
			return experts + amateurs;
		}
		
		public int compareTo(MPRegionDataItem inst)
		{
			int result = region.compareTo(inst.region);
			return result;
		}
		
	}
		
	ArrayList<MPRegionDataItem> regiondatalist = null;
	
	public MPRegionDataList()
	{
		regiondatalist = new ArrayList<MPRegionDataItem>();
	}
	
	public void AddQuad(String region, int quad)
	{
		MPRegionDataItem item = new MPRegionDataItem(region, quad, 0, 0);
		Add(item);
	}
	
	public void AddExpert(String region)
	{
		MPRegionDataItem item = new MPRegionDataItem(region, 0, 1, 0);
		Add(item);
	}
	
	public void AddAmateur(String region)
	{
		MPRegionDataItem item = new MPRegionDataItem(region, 0, 0, 1);
		Add(item);		
	}
	
	public void AddAveNoteDur(String region, double dur)
	{
		MPRegionDataItem item = new MPRegionDataItem(region, 0, 0, 0);
		item.notedur = dur;
		item.notecount = 1;
		Add(item);
	}
	
	public void Add(MPRegionDataItem item)
	{
		int index = Collections.binarySearch(regiondatalist, item);
		if (index >= 0) {
			MPRegionDataItem match = regiondatalist.get(index);
			match.Add(item);
		} else {
			regiondatalist.add(-index - 1, item);
		}
	}

	public void Merge(MPRegionDataList list2)
	{
		for (MPRegionDataItem item : list2.regiondatalist) {
			Add(item);
		}
	}
	
	public int GetSize()
	{
		return regiondatalist.size();
	}
	
	public MPRegionDataItem GetRegionData(String region)
	{
		MPRegionDataItem key = new MPRegionDataItem(region);
		int index = Collections.binarySearch(regiondatalist, key);
		if (index < 0) {
			return null;
		}
		
		return regiondatalist.get(index);
	}
	
}
