import java.util.*;

public class MPRegionPairList {
	private ArrayList<MPRegionPair> regions;
	
	public MPRegionPairList()
	{
		regions = new ArrayList<MPRegionPair>();
	}
	
	public void Add(MPRegionPair rp)
	{
		// System.out.println("new pair: " + rp.region1 + " " + rp.region2);
		regions.add(rp);
	}
	
	public int Size()
	{
		return regions.size();
	}
	
	public void Sort()
	{
		Collections.sort(regions, new MPRegionPairComparator());
	}
	
	public void SortByPairValue()
	{
		Collections.sort(regions, new MPRegionPairValueComparator());
	}
	
	public void SoryByRegion1()
	{
		Collections.sort(regions, new MPRegionPair1Comparator());
	}
	
	public MPRegionPair Find(String key)
	{
		for (MPRegionPair rp : regions) {
			if (rp.region1.GetRegionName().compareTo(key) == 0) {
				return rp;
			}
		}
		return null;
	}
	
	public void Merge(MPRegionPairList list)
	{
		if (list != null) {
			for (MPRegionPair rp : list.regions) {
				regions.add(rp);
			}
		}
	}
	
	public ArrayList<RIRegionItem> RegionListFromRegions()
	{
		// build a list of all the unique regions
		TreeMap<String, RIRegionItem> regiontree = new TreeMap<String, RIRegionItem>();
		
		for (MPRegionPair rp : regions) {
			regiontree.put(rp.region1.GetRegionName(), rp.region1);
			regiontree.put(rp.region2.GetRegionName(), rp.region2);
		}
		
		ArrayList<RIRegionItem> regionlist = new ArrayList<RIRegionItem>(regiontree.values());

		/*
		Collection<RIRegionItem> c = (Collection<RIRegionItem>)regiontree.values();
		Iterator<RIRegionItem> sitr = c.iterator();
		ArrayList<RIRegionItem> regionlist = new ArrayList<RIRegionItem>();
	    while (sitr.hasNext()) {
	    	RIRegionItem item = (RIRegionItem)sitr.next();
	    	regionlist.add(item);
	    }
	    */
	    
	    regiontree.clear();
		return regionlist;
	}
	
	public MPRegionPair AggregateFinalRegionPair(RIRegionItem region)
	{
		long cum_pairs = 0;
		int c1 = 0;
		int hits = 0;
		double pearson = 0;
		String region_name = region.GetRegionName();
		for (MPRegionPair rp : regions) {
			boolean r1match = rp.region1.GetRegionName().compareTo(region_name) == 0;
			if (r1match || rp.region2.GetRegionName().compareTo(region_name) == 0) {
				if (r1match) {
					c1 += rp.region1size;
				} else {
					c1 += rp.region2size;
				}
				long r1r2 = (long)rp.region1size * rp.region2size;
				cum_pairs += r1r2;
				pearson += rp.pearson * r1r2;
				++hits;
			}
		}
		
		if (hits > 0) {
			pearson = pearson / (double)cum_pairs;
			MPRegionPair rp = new MPRegionPair(region, c1, null, hits, (float)pearson);
			return rp;
		} 
		
		return null;
	}
	
	public MPRegionPairList AggregateFinalRegionPairList(ArrayList<RIRegionItem> theregions)
	{
		MPRegionPairList final_external_list = new MPRegionPairList();
		for (RIRegionItem s : theregions) {
			MPRegionPair rp = AggregateFinalRegionPair(s);
			if (rp != null) {
				final_external_list.Add(rp);
			}
		}
		return final_external_list;
		
	}
	
	/***
	 * 
	 * Walk the region pair list and merge all country/country pairs into a single pair, 
	 * returning the new (sorted) list
	 * 
	 * @param min_region_size:  how many matching pairs per region is the minimum
	 * @return the newlist
	 */
	public MPRegionPairList AggregateLikePairs(Comparator<MPRegionPair> compFunction)
	{
		MPRegionPairList newlist = new MPRegionPairList();
		
		int i = 0;
		while (i < regions.size()) {
			int j = i + 1;
			int c1 = regions.get(i).region1size;
			int c2 = regions.get(i).region2size;
			long product = c1 * c2;
			double pearson = regions.get(i).pearson * product;
			boolean match = true;
			while (match && j < regions.size()) {
				match = compFunction.compare(regions.get(i), regions.get(j)) == 0;
				if (match) {
					int cc1 = regions.get(j).region1size;
					int cc2 = regions.get(j).region2size;
					c1 += cc1;
					c2 += cc2;
					long cc1cc2 = cc1 * cc2;
					pearson += regions.get(j).pearson * cc1cc2;
					product += cc1cc2;
					++j;
				}
			}

			// if (c1 > 1 && c2 > 1) {
				pearson = pearson / (product);
				MPRegionPair rp = new MPRegionPair(regions.get(i).region1, c1, regions.get(i).region2, c2, (float)pearson);
				newlist.Add(rp);
			//}
			
			i = j;
		}
		
		return newlist;
	}
	
	/**
	 * 
	 * walk a list of region pairs which includes country V country as well as country & country compares
	 * 
	 * return a list of the country & country compares; prune these from the list
	 * 
	 */
	private MPRegionPairList CollectInternalCompares()
	{
		MPRegionPairList internal_list = new MPRegionPairList();
		
		int i = 0;
		while (i < regions.size()) {
			MPRegionPair apair = regions.get(i);
			String r1 = apair.region1.GetRegionName();
			String r2 = apair.region2.GetRegionName();
			if (r1.matches(r2)) {
				regions.remove(i);
				internal_list.regions.add(apair);
			} else {
				++i;
			}
		}
		
		return internal_list;
	}
	
	private void finalReport(MPOutputStream out, MPUniqueIDs ids, MPRegionDataList rdlist, int min_uniques, 
			MPRegionPairList final_external_list, MPRegionPairList final_internal_list)
	{
		out.MPWrite("region\trank");
		out.MPWrite("\tcross pearson\tinternal pearson\tN\tregions\tperformances\tuniques\t");
		out.MPWrite("urban%\t\texpert%\t\ttime\t\t");
		// out.MPWriteLn("slow slow\tslow fast\tfast slow\tfast fast\n");
		out.MPWrite("\n\n");
		System.out.println("Sort 2");
		final_external_list.SortByPairValue();
		int rank = 1;
		System.out.print("Loop " + final_external_list.Size() + "\n" );
		
		for (MPRegionPair rp : final_external_list.regions) {
			MPUniqueIDs.MPUniqueIDInfo info = ids.GetUniqueIDInfoCountByRegion(rp.region1);
			int uniques_in_region = info.GetRegionCnt();
			if (uniques_in_region >= min_uniques) {
				float urban = (float)info.GetUrbanCount() / (float)(uniques_in_region);
				MPRegionDataList.MPRegionDataItem regiondata = rdlist.GetRegionData(rp.region1.GetRegionName());
				int expertcnt = regiondata.getExperts();
				int totalcnt = regiondata.getTotal();

				MPRegionPair crp = final_internal_list.Find(rp.region1.GetRegionName());
				float internal_pearson = 0f;
				int crp_size = 0;
				if (crp != null) {
					internal_pearson = crp.pearson;
					crp_size = crp.region1size;
				}

				out.MPWrite(info.region_full_name + "\t" + rank++ + "\t");
				out.MPWrite(rp.pearson + "\t" + internal_pearson + "\t");
				out.MPWrite(rp.region1size + "\t" + rp.region2size + "\t" + crp_size + "\t");

				out.MPWrite(uniques_in_region + "\t");
				out.MPWrite(urban + "\t\t");
				out.MPWrite((float)expertcnt / (float)totalcnt + "\t\t");

				out.MPWrite(regiondata.notedur / (double)regiondata.notecount + "\t\t");
				// MPRegionDataList.Quads quads = regiondata.getQuads();
				// quads.Report(out);

				out.MPWriteLn("");
			}			
		}
	}
		
	@SuppressWarnings("unused")
	private void finalRegionPairReport(MPOutputStream out, MPUniqueIDs ids, int min_uniques, MPRegionPairList external_list)
	{
		// this list is, unlike final_external_list, non-aggregated:  it contains each pair
		System.out.println("Final region pairs");
		out.MPWriteLn("\nregion pairs");
		for (MPRegionPair rp : external_list.regions) {
			if (rp.region1size >= min_uniques && rp.region2size >= min_uniques) {
				String s = ids.getRegionFullName(rp.region1.GetRegionName()) + "\t" + rp.region1size + "\t";
				s = s + ids.getRegionFullName(rp.region2.GetRegionName()) + "\t" + rp.region2size + "\t";
				s = s + rp.pearson;
				out.MPWriteLn(s);
			}
		}		
	}
		
	
	/*
	 * 
	 * This function this function aggregates like region pairs which were derived from different songs.
	 * 
	 * It builds three lists of regions: 
	 * 
	 *  external_list - a list of all region 2 region external compares
	 *  final_internal_list - a list of internal region compares
	 *  final_external_list - a list of external region compares (aggregation of external_list)
	 * 
	 * It then reports these lists via the 'out' parameter.
	 */
	
	public void AggregateAndReport(MPOutputStream out, MPUniqueIDs ids, MPRegionDataList rdlist, int min_uniques)
	{
		System.out.println("sort");
		Sort();
		
		// aggregate compares between like regions
		// independent of whether the compare is to the same country (r1 == r2) or across (r1 != r2)
		// in the process, internal country compares have been combined
		MPRegionPairList aggregated_list;
		System.out.println("aggregate like pairs");
		aggregated_list = AggregateLikePairs(new MPRegionPairComparator());
		
		// next, extract all pairs that are the same country (internal comparison),
		// pruning those pairs from the aggregated list in the process
		MPRegionPairList final_internal_list;
		System.out.println("collect internal compares");
		final_internal_list = aggregated_list.CollectInternalCompares();
		
		// the internal compares have been removed from the aggregated list
		// this is now the external list of compares
		MPRegionPairList external_list = aggregated_list;
		aggregated_list = null;
		
		// now create a list of regions from this external list
		System.out.println("region list from regions");
		external_list.SortByPairValue();
		ArrayList<RIRegionItem> theregions = external_list.RegionListFromRegions();
		
		// aggregate external compares into final list
		System.out.println("aggregate final region pairs");
		MPRegionPairList final_external_list = external_list.AggregateFinalRegionPairList(theregions);
				
		finalReport(out, ids, rdlist, min_uniques, final_external_list, final_internal_list);
		// finalRegionPairReport(out, ids, min_uniques, external_list);
	}
	
	public void ReportByCountryX(MPOutputStream out, MPUniqueIDs ids, MPPerfQuadrants perfQuads, int min_region_size)
	{
		Sort();
		
		out.MPWriteLn("\nRank by Country\t" + ids.GetSize());
		MPRegionPairList mpcountrylist;
		mpcountrylist = AggregateLikePairs(new MPRegionPairComparator());
		mpcountrylist.SortByPairValue();

		for (MPRegionPair crp : mpcountrylist.regions) {
			MPUniqueIDs.MPUniqueIDInfo info = ids.GetUniqueIDInfoCountByRegion(crp.region1);
			float rcount = info.GetRegionCnt();
			float urban =  (float)info.GetUrbanCount() / (float)(rcount);
			out.MPWrite(crp.region1 + "\t" + crp.region1size + "\t" + crp.pearson + "\t" + Math.round(rcount) + "\t" + urban + "\t\t");
			
			MPPerfQuadrants.Quads quads = perfQuads.GetQuadInfo(crp.region1.GetRegionName());
			quads.Report(out);
			
			out.MPWriteLn("");
		}
	
	}
}
