import java.sql.*;
import java.util.ArrayList;

/** 
 * MPSQL
 * This class is used to interface with our MySQL tables where
 * performances and their associated resource UUIDs are stored.
 * 
 * @author Nick Rudolfsky & jeffreysmith
 * @date Summer 2012
 * @version 1.0
 */

public class MPSQL 
{
	/** Singleton instance of connection to our MySQL database */
	private static Connection sqlConnection = null;
	
	/** Singleton instance of MySQL statment (for making queries) */
	private static Statement sqlStatement = null;
	
	/** Debug flag for logging */
	private static boolean DEBUG_SQL = false;
	
	/**
	 * Connect to our MySQL database
	 * @return true if connected successfully, false otherwise
	 */
	public static boolean connectToSQL()
	{
        return false;
	}
	
	/**
	 * Execute a MySQL query
	 * @param query A MySQL query
	 * @return the result set returned by the query
	 */
	public static ResultSet executeQuery(String query)
	{
		if (DEBUG_SQL)
			System.out.println("Executing MySQL query: " + query);
		
		ResultSet result = null;
		
		try {
			// execute the query
        	result = sqlStatement.executeQuery(query);

        } catch (SQLException e) {
           printSQLException(e, "connectToSQL");
        }
		
		return result;
	}
	
	public static class PerfData
	{
		String perf_uuid;
		int user_uuid;
		Date perf_date;		
		
		public PerfData(String s, int i, Date d)
		{
			perf_uuid = s;
			user_uuid = i;
			perf_date = d;
		}
	}
	
	/**
	 * Get Magic Piano performances from our MySQL server
	 * @param song_uid The uid of the desired track. Can be obtained from Smule Admin Tool
	 * @param numPerf The number of performances to return
	 * @return A string array of MPPerformance meta data objects which contain resourceUUID of file to download from AWS
	 */
	public static ArrayList<MPPerformance.MPMetaData> getPerformancesForSong(String song_uid, int numPerfs)
	{
		MPDialogs.MPProgressDialog.showProgressDialog("Finding Performances For " + song_uid, true);

		System.out.println("Downloading " + numPerfs + " performance for song " + song_uid);
		ArrayList<MPPerformance.MPMetaData> performances = new ArrayList<MPPerformance.MPMetaData>();
		
		// execute SQL query to get desired Magic Piano performances
/*		String query = "SELECT uuid FROM performances " + "WHERE song_uid = '" + song_uid + "' AND " +
						"app_uid = 'minipiano' AND is_rendered = 1 AND " + "format = 'midi' AND " +
						"app_version != '4.0.0' " + "LIMIT " + numPerfs;
*/

/*
		String query = "SELECT uuid, user_id, created_at FROM performances " + "WHERE song_uid = '" + song_uid + "' AND " +
				"app_uid = 'minipiano' AND is_rendered = 1 AND " + "format = 'midi' AND " +
				"app_version != '4.0.0' ORDER BY user_id ASC " + "LIMIT " + numPerfs;
*/


 		String query = "SELECT uuid, user_id, created_at FROM performances " + "WHERE song_uid = '" + song_uid + "' AND " +
						"app_uid = 'minipiano' AND is_rendered = 1 AND " + "format = 'midi' AND " +
						"app_version != '4.0.0' " + "LIMIT " + numPerfs;

		
/* this query only pulls performances after a given date 
		String query = "SELECT uuid, user_id, created_at FROM performances " + "WHERE song_uid = '" + song_uid + "' AND " +
				"app_uid = 'minipiano' AND is_rendered = 1 AND " + "format = 'midi' AND " +
				"app_version != '4.0.0' " + 
				" AND created_at >= '2013-02-03 00:00:00'" + 
				"LIMIT " + numPerfs;
*/


		System.out.println(query);
		ResultSet perfResults = executeQuery(query);
		ArrayList<PerfData> perf_data_list = new ArrayList<PerfData>();
		
		try {
			// copy the resulting performance uuids in an array
			
			int cnt = 0;
			while ( cnt < numPerfs && perfResults != null && perfResults.next() ) {
				String perfid = perfResults.getString(1);
				int userid = perfResults.getInt("user_id");
				Date perfdate = perfResults.getDate("created_at");
				PerfData pdata = new PerfData(perfid, userid, perfdate);
				perf_data_list.add(pdata);
				++cnt;
			}

			// performance result set is not longer needed. close.
			if (perfResults != null) { 
				perfResults.close(); 
			}

			MPDialogs.MPProgressDialog.showProgressDialog("Locating Performance Resources", false);

			// get track for each performance
			int perfNum = 0;
			for (PerfData perf_data : perf_data_list) {
				// update our progress dialog
				perfNum++;
				MPDialogs.MPProgressDialog.updateProgress((float)perfNum / (float)numPerfs);
				
				// execute query to get track_uuid
				query = "SELECT track_uuid FROM performances_tracks WHERE perf_uuid = '" + perf_data.perf_uuid + "'";
				ResultSet trackResult = MPSQL.executeQuery(query);

				// there should only be on track returned from above query...
				if ( trackResult != null && trackResult.next() ) {
					String track_uuid = trackResult.getString(1);
					MPPerformance.MPMetaData performanceData = null;

					// execute a query to get the resource for this track
					query = "SELECT resource_uuid FROM resources_tracks WHERE track_uuid = '" + track_uuid + "'";
					ResultSet resourceResult = MPSQL.executeQuery(query);
					
					// get the resource for this track
					if ( resourceResult != null && resourceResult.next() ) {
						String resource = resourceResult.getString(1) + ".midi";

						// this is the resource we can download from S3. add to array
						performanceData = new MPPerformance.MPMetaData(resource, perf_data.user_uuid, perf_data.perf_date);
					}

					// close resource result
					if (resourceResult != null) { 
						resourceResult.close(); 
					}
					
					// download location meta data
					query = "SELECT latitude, longitude FROM tracks WHERE uuid = '" + track_uuid + "'";
					ResultSet metaResult = MPSQL.executeQuery(query);
					
					// get the resource for this track
					if ( metaResult != null && metaResult.next() ) {
						// set performance meta data
						performanceData.latitude = metaResult.getString(1);
						performanceData.longitude = metaResult.getString(2);
					}
					
					// close meta result
					if (metaResult != null) { 
						metaResult.close(); 
					}
					
					// add the performance to the array
					if (performanceData != null) {
						performances.add(performanceData);
					}
				}

				// close track result
				if (trackResult != null) { 
					trackResult.close(); 
				}
			}
		} catch (SQLException e) {
			printSQLException(e, "getPerformancesForSong");
		}
		
		return performances;
	}
	
	/**
	 * Helper method used to log an SQL exception
	 * @param e the exception to print
	 */
	public static void printSQLException(Exception e, String methodName)
	{
		System.out.println(methodName + " exception:");
		
		if (e instanceof SQLException)
		{
			SQLException sqle = (SQLException)e;
			System.out.println("Error Message:    " + sqle.getMessage());
            System.out.println("SQL Error Code:   " + sqle.getErrorCode());
            System.out.println("SQL State: " 		+ sqle.getSQLState());
		}
		else
		{
			System.out.println("Exception: " + e);
		}
	}
}
