import java.io.File;
import java.util.ArrayList;

public class MPSongIO {

	protected String songFileName;
	private MPJson json;

	public void initMeta(MPSongList perfList, File dir, int perf_max)
	{
		String curPath = dir.getPath();
		songFileName = dir.getName(); 

		loadJSON();
		
		String[] directoryItems = dir.list();

		// Iterate over all songs in directory; assume no nested directories
		for (int i = 0; i < directoryItems.length && perfList.getSize() < perf_max; i++) {
			String directoryItem = directoryItems[i];
						
			// determine path for new item
			String newPath;
			if (curPath.equals(".")) {
				newPath = directoryItem;
			} else {
				newPath = curPath + File.separator + directoryItem;
			}

			// determine if directory or file
			File fileItem = new File(newPath);
			if (!fileItem.isDirectory() && !fileItem.isHidden()) {
				if (fileItem.getName().endsWith("midi")) {
					MPPerformance thePerf = new MPPerformance(this, fileItem);
					perfList.Add(thePerf);
				}
			}
		}
	}
	
	public MPSongIO(MPSongList perfList, File dir, int perf_max)
	{
		initMeta(perfList, dir, perf_max);
	}
	
	private void loadJSON()
	{
		String curPath = "orgperfs" + File.separator + songFileName + ".json";
		json = new MPJson();
		if (!json.MPJsonOpenFile(curPath)) {
			json = null;
		}
	}
	
	public float GetSongJSONRate()
	{
		if (json == null) {
			return (float)1.0;
		}
		
		double d = json.MPJsonGetDouble("firefly_speed");
		// assume 4.0 is 1.0 for firefly rate
		float rate = (float)(d / 4.0);
		return rate;
	}
	
	public MPOriginalPerformance LoadOrgPerf()
	{
		// now load the original midi performance (the score)
		String curPath = "orgperfs" + File.separator + songFileName + ".midi";
		File fItem = new File(curPath);
		if (!fItem.isFile()) {
			System.err.println("failed to load org perf");
			return null;
		}
		
		// original performance constructor does no analysis
		float rate = GetSongJSONRate();
		MPOriginalPerformance p = new MPOriginalPerformance(this, fItem, rate);
		return p;
	}

	public boolean useChords()
	{
		boolean b = json.MPJsonGetBoolean("use_chords");
		return b;
	}
	
	public int[] getChordIndex()
	{
		int intarray[] = json.MPJsonGetIntArray("chords");
		return intarray;
	}
	

	public MPPhrase getPhraseIndex(String sindex)
	{
		if (json == null) {
			return null;
		}
		
		ArrayList<Integer> alist = json.MPJsonGetIntList(sindex);
		if (alist == null) {
			return null;
		}
		
		MPPhrase index = new MPPhrase(alist.size() - 1);
		for (int i = 0; i < alist.size(); i++) {
			index.phrase[i] = alist.get(i);
		}
		
		return index;
	}
	
	public String GetSongFileName()
	{
		return songFileName;
	}
	

}
