import java.util.ArrayList;
import java.util.Collections;
import java.util.GregorianCalendar;
import java.util.List;
import java.util.TreeSet;

/***
 * 
 * @author jeffreyniusmith
 * 
 * Nov 7, 2012
 * 
 * Class to manipulate SongList (a list of Songs)
 *
 */
public class MPSongList {
	
	public ArrayList<MPPerformance> songs;

	public MPSongList()
	{
		songs = new ArrayList<MPPerformance>();
	}
	
	public MPSongList(List<MPPerformance> list) 
	{
		songs = new ArrayList<MPPerformance>(list);
	}
	
	protected void Clear()
	{
		for (MPPerformance p : songs ) {
			p.Clear();
		}
		songs.clear();
	}
	
	protected int getSize()
	{
		return songs.size();
	}
	
	protected void Add(MPPerformance perf)
	{
		songs.add(perf);
	}
	
	protected MPPerformance Get(int i)
	{
		return songs.get(i);
	}
	
	/***
	 * 
	 * @param a,b -- two performance lists 
	 * 
	 * @return the number of performances that exist on both lists
	 */
	protected int perfIntersectionCnt(MPSongList b) 
	{
		int count = 0;
		for (MPPerformance q : songs) {
			if (b.songs.contains(q)) {
				++count;
			}
		}
		return count;
	}
		
	protected MPRegionDataList fetchRegionData(MPSongStats stats, MPPhrase phrase, int region_type)
	{
		MPRegionDataList rdlist = new MPRegionDataList();
		for (MPPerformance q : songs) {
			RIRegionItem item = RIRegionItem.AllocRegionItem(region_type, q.getLoc());
			if (item != null) {
				String regionname = item.GetRegionName();
				if (regionname != null) {
					// rdlist.AddQuad(regionname, q.getSlopeQuadrant(stats));
					if (phrase != null) {
						if (q.PhraseMatch(phrase, stats)) {
							rdlist.AddExpert(regionname);
						} else {
							rdlist.AddAmateur(regionname);
						}
					}
					double d = (double)q.noteAverageDuration / (double)stats.averageNoteLength;
					rdlist.AddAveNoteDur(regionname, d);
				}
			}
		}
		
		return rdlist;
	}
	
	protected MPPerfQuadrants fetchPerfQuads(MPSongStats stats)
	{
		MPPerfQuadrants quads = new MPPerfQuadrants();
		
		for (MPPerformance q : songs) {
			if (q.perfLoc.getCountry() != null) {
				quads.Add(q.perfLoc.getCountry(), q.getSlopeQuadrant(stats));
			}
		}
		
		return quads;
	}
	
	protected int GetExperts(MPPhrase phrase, MPSongStats stats)
	{
		int expert_count = 0;
		
		if (phrase != null) {
			for (MPPerformance q : songs) {
				if (q.perfLoc.getCountry() != null) {
					if (q.PhraseMatch(phrase, stats)) {
						++expert_count;
					}
				}
			}
		}
		
		return expert_count;
	}
	
	protected float urbanPercent()
	{
		int count = 0;
		int urban = 0;
		for (MPPerformance p : songs) {
			if (!p.perfLoc.isGeoCodeNull()) {
				if (p.perfLoc.isCity()) {
					++urban;
				}
				++count;
			}
		}
		return (float)urban / (float)count;
	}
	
	protected void Sort()
	{
		Collections.sort(songs);		
	}
	
	protected void SortByValue()
	{
		Collections.sort(songs, new MPPerformanceNoteValueComparator());
	}
	
	protected void SortByPearson()
	{
		Collections.sort(songs, new MPPerformanceNoteValueComparator());
		
	}
	
	public void pruneChords(MPOriginalPerformance org_perf)
	{
		if (org_perf == null) {
			System.err.println("prune chords:  no org perf loaded");
			return;
		}
		
		short cnt_c[] = {0, 0, 0, 0, 0, 0};
		short cnt_j[] = {0, 0, 0, 0, 0, 0};

		ArrayList<MPPerformance> newList = new ArrayList<MPPerformance>();

		for (MPPerformance p : songs) {
			boolean uses_chords = true;
			for (int i = 0; uses_chords && i < p.getNumNotes(); i++) {
				
				// hack?:  this should be zero, but i'm using perfs as subs for org perf
				if (org_perf.getNoteDur(i) == 0) {
					
					// magic piano mechanic gives 0 or 1 for chord notes
					if (p.getNoteDur(i) > 1) {
						uses_chords = false;
					}
				}
			}
			if (!uses_chords) {
				newList.add(p);
			} else {
				cnt_c[p.perfLoc.geoIndex()]++;
			}
			cnt_j[p.perfLoc.geoIndex()]++;
		}
		
		for (int i = 0; i < 6; i++) {
			System.out.println(MPGeoCoord2.geoIndexToString(i) + " " + (cnt_c[i] / (float)cnt_j[i]));
		}
		
		int total_prune = songs.size() - newList.size();
		System.out.println("pruned x chords " + total_prune + " " + (total_prune/(float)songs.size()) + "%");
		songs.clear();
		songs = newList;
	}
	
	public void pruneNonGeoData()
	{
		int i = 0;
		while (i < songs.size()) {
			MPPerformance p = songs.get(i);
			if (!p.perfLoc.isGeoCodeNull()) {
				++i;
			} else {
//				System.out.println("Nuking (d/lat) " + aPerformance.averageDuration + " " + 
//						aPerformance.n_lat + " " + aPerformance.performanceFile.getPath());
				songs.remove(i);
				p.Clear();
			}
		}
	}
	
	protected boolean pruneHighStdDevPerfs(MPSongStats stats)
	{
		
		int i = 0;
		int arraymax = stats.maxNoteCount + 1;
		int songcount[] = new int[arraymax];
		int maxnotes = 0;
		while (i < songs.size()) {
			MPPerformance p = songs.get(i);
			int cnt = p.getNumNotes();
			if (cnt > maxnotes) {
				maxnotes = cnt;
			}
			if (cnt > arraymax) {
				System.err.println("bogus note count " + p.performanceFile.getPath());
			}
			songcount[cnt]++;
			++i;
		}
		
		// look for a bucket > 1/2 of total performances
		int notecount = 0;
		int half = i / 2;
		for (int j = 0; j <= maxnotes; j++) {
			if (songcount[j] > half) {
				notecount = j;
				break;
			}
		}
		
		if (notecount == 0) {
			System.err.println("could not find median note length");
			return false;
		}
				
		// prune any song that has non-standard note count
		ArrayList<MPPerformance> new_list = new ArrayList<MPPerformance>(songs.size());
		for (int j = 0; j < songs.size(); j++) {
			MPPerformance p = songs.get(j);
			if (p.getNumNotes() != notecount) {
				// System.out.println("bad note count, nuking " + p.performanceFile.getPath());
				// p.PermanentlyDeleteFile();
				// songs.remove(j);
				p.Clear();
			} else if (p.noteAverageDuration > stats.averageNoteLength * 2 ||
						p.noteAverageDuration < stats.averageNoteLength / 2) {
				// songs.remove(j);
				p.Clear();
			} else {
				new_list.add(p);
			}
		}
		
		// work around java bug with delete?  
		if (new_list.size() != songs.size()) {
			songs.clear();
			songs = new_list;
			return true;
		}
		
		return false;
	}

	protected void pruneUrban()
	{
		int i = 0;
		int old_size = songs.size();
		while (i < songs.size()) {
			MPPerformance p = songs.get(i);
			if (p.perfLoc.isCity()) {
				songs.remove(i);
				p.Clear();
			} else {
				++i;
			}
		}
		System.out.println("prune cities " + (old_size - songs.size()));
	}
	
	protected void pruneRural()
	{
		int i = 0;
		int old_size = songs.size();
		while (i < songs.size()) {
			MPPerformance p = songs.get(i);
			if (!p.perfLoc.isCity()) {
				songs.remove(i);
				p.Clear();
			} else {
				++i;
			}
		}
		System.out.println("prune rural " + (old_size - songs.size()));
	}
	
/****
	 * Remove list elements newer than recentDate
	 */
	protected void pruneLeastRecent(int year, int month, int date)
	{
		GregorianCalendar thedate = new GregorianCalendar();
		
		// month zero based
		thedate.set(year, month - 1, date);
		MPPerfDate recentDate = new MPPerfDate(thedate);
		System.out.println("recent date " + recentDate.getString());
		ArrayList<MPPerformance> newList = new ArrayList<MPPerformance>();
		
		int i = 0;
		while (i < songs.size()) {
			MPPerformance p = songs.get(i);
			if (p.perfDate.compareTo(recentDate) > 0) {
				// System.out.println("keep " + aPerformance.perfDate.getString());
				newList.add(p);
				++i;
			} else {
				System.out.println("prune least recent " + p.perfDate.getString());
				// songs.remove(i);
				p.Clear();
			}
		}
		System.out.println("prune least recent " + i);
		songs.clear();
		songs = newList;
	}	
	
	/****
	 * Remove list elements older than recentDate
	 */
	protected void pruneMoreRecent(int year, int month, int date)
	{
		GregorianCalendar thedate = new GregorianCalendar();
		
		// month zero based
		thedate.set(year, month - 1, date);
		MPPerfDate recentDate = new MPPerfDate(thedate);
		// System.out.println("recent date " + recentDate.getString());
		ArrayList<MPPerformance> newList = new ArrayList<MPPerformance>();
		
		int i = 0;
		int pruned = 0;
		for (MPPerformance p : songs) {
			if (p.perfDate.compareTo(recentDate) <= 0) {
				newList.add(p);
				// System.out.println("keep " + p.perfDate.getString());
			} else {
				++pruned;
				System.out.println("more recent prune " + p.perfDate.getString());
				p.PermanentlyDeleteFile();
			}
		}
		if (pruned > 0) {
			System.out.println("prune more recent " + pruned + " " + i);
		}
		songs.clear();
		songs = newList;
	}	
	
	protected void pruneRepeatPerformances()
	{
//		Collections.sort(songs, new MPPerformanceDateComparator());
		Collections.sort(songs);
		ArrayList<MPPerformance> newList = new ArrayList<MPPerformance>();

		int i = 0; 
		while (i < songs.size() - 1) {
			MPPerformance a = songs.get(i);
			int j = i + 1;
			boolean repeat = true;
			while (repeat && j < songs.size()) {
				MPPerformance b = songs.get(j);
				if (a.getUserID() == b.getUserID()) {
					// songs.remove(i);
					b.Clear();
					++j;
				} else {
					newList.add(b);
					repeat = false;
				}
			}
			i = j;
		}		
		
		System.out.println("total repeat count " + (songs.size() - newList.size()) + " " + songs.size()); 

		songs.clear();
		songs = newList;
	}
	
	/***
	 * 
	 * prune all first performances by a user (which yields a list of only repeat performances)
	 * 
	 */
	
	protected void pruneFirstPerformances()
	{
//		Collections.sort(songs, new MPPerformanceDateComparator());
		Collections.sort(songs);
		
		ArrayList<MPPerformance> newList = new ArrayList<MPPerformance>();

		int i = 0; 
		while (i < songs.size() - 1) {
			MPPerformance a = songs.get(i);
			int j = i + 1;
			boolean repeat = true;
			while (repeat && j < songs.size()) {
				MPPerformance b = songs.get(j);
				if (a.getUserID() == b.getUserID()) {
					newList.add(b);
					++j;
				} else {
					repeat = false;
				}
			}
			i = j;
		}		
		
		songs.clear();
		songs = newList;

//		System.out.println("total repeat count " + repeatcount + " " + songs.size()); 

	}
	
	/***
	 * 
	 * all caching logic for clusters, etc. will use the unique performance ID, which
	 * the caching logic assumes must be 0..size + 1
	 * 
	 * given we prune several list items, this performance ID needs to be reset 
	 * 
	 * obviously this reset must occur before we start caching
	 * 
	 */
	
	public boolean perfExists(MPPerformance p)
	{
		return songs.contains(p);
	}

	protected MPSongList getSongsInRegion(RIRegionItem region)
	{
		MPSongList the_list = new MPSongList();
		for (MPPerformance perf : songs) {
			if (region.Contains(perf.perfLoc)) {
				the_list.Add(perf);
			} 
		}
		return the_list;
	}
	
	/*
	 * build a list of region items that from the songlist
	 * 
	 * this list will be a subset of the region_items parameter, a tree of all region items in all songs
	 */
	protected ArrayList<RIRegionItem> RegionListFromPerfList2(int region_type, RIRegionItemSet region_items)
	{
		TreeSet<RIRegionItem> rgnlist = new TreeSet<RIRegionItem>();
		for (MPPerformance perf : songs) {
			if (!perf.perfLoc.isGeoCodeNull()) {
				String key = RIRegionItem.GetRegionKey(region_type, perf.perfLoc);
				if (key != null) {
					
					// does this region item exist, yet, in the tree?
					RIRegionItem item = region_items.Get(key);
					
					// if not, add this item to tree
					if (item == null) {
						item = RIRegionItem.AllocRegionItem(region_type, perf.perfLoc);
						region_items.Put(item);
					}
					
					rgnlist.add(item);
				}
			}
		}

		return new ArrayList<RIRegionItem>(rgnlist);
	}
	
	/*
	 * build a list of region items that from the songlist
	 * 
	 * this list will be a subset of the region_items parameter, a tree of all region items in all songs
	 */
	protected ArrayList<RIRegionItem> RegionListFromPerfList(int region_type)
	{
		TreeSet<RIRegionItem> rgnlist = new TreeSet<RIRegionItem>();
		for (MPPerformance perf : songs) {
			String key = RIRegionItem.GetRegionKey(region_type, perf.perfLoc);
			if (key != null) {
				RIRegionItem item = RIRegionItem.AllocRegionItem(region_type, perf.perfLoc);
				rgnlist.add(item);
			}
		}
		return new ArrayList<RIRegionItem>(rgnlist);
	}
	
	protected ArrayList<RIRegionItem> CountryListFromPerfList()
	{
		return RegionListFromPerfList(MPConfig.ANALYZE_BY_COUNTRY);
	}
	
	protected MPSongList getPerfsFromCountry(String country)
	{
		MPSongList clist = new MPSongList();
		for (MPPerformance perf : songs) {
			if (!perf.perfLoc.isGeoCodeNull()) {
				if (country.compareTo(perf.perfLoc.getCountry()) == 0) {
					clist.Add(perf);
				}
			}
		}
		return clist;
	}
	

}
