import java.util.ArrayList;

public class MPSongStats {
	
	public ArrayList<Float> aveNoteDur;
	public ArrayList<Float> stdDevNoteDur;
	public ArrayList<Float> aveNoteSlopeX;
	
	class ChordInfo {
		short key;
		int value;
		boolean is_chord;
		boolean boundary;
		
		public ChordInfo(short k, int v, boolean c)
		{
			key = k;
			value = v;
			is_chord = c;
		}
	}
	int num_performances;
	public ArrayList<ChordInfo> chords;

	float averageNoteLength;
	float stdDevNoteLength;
	float repeatSongPercent;
	short maxNoteCount;
	
	public void Clear()
	{
		if (aveNoteDur != null) {
			aveNoteDur.clear();
		}
		if (stdDevNoteDur != null) {
			stdDevNoteDur.clear();
		}
		if (aveNoteSlopeX != null) {
			aveNoteSlopeX.clear();
		}
		if (chords != null) {
			chords.clear();
		}
	}
	
	public void computeSongStats(MPSongList sngList)
	{
		int cnt = 0;
		float stddev = 0;
		float average = 0;
		for (cnt = 0; cnt < sngList.getSize(); cnt++) {
			MPPerformance aPerformance = sngList.Get(cnt);
			average += aPerformance.noteAverageDuration;
			stddev += aPerformance.noteAverageDuration * aPerformance.noteAverageDuration;
			if (aPerformance.getNumNotes() > maxNoteCount) {
				short s = (short)aPerformance.getNumNotes();
				maxNoteCount = s;
			}
		}
						
		average = average / cnt;
		averageNoteLength = average;
		
		if (cnt > 1) {
			average = average * average;	
			float a2 = average * ((float)cnt / (cnt - 1));
			float s2 = stddev / ((float)cnt - 1);
			stdDevNoteLength = (float)Math.sqrt(s2 - a2);
		} else {
			stdDevNoteLength = 0;
		}
	}
	
	private void computeNoteStats(MPSongList sngList)
	{		
		if (maxNoteCount == 0) {
			System.out.println("computeNoteStats: error:  no song stats"); 
			return;
		}
		
		aveNoteDur = new ArrayList<Float>(maxNoteCount);
		stdDevNoteDur = new ArrayList<Float>(maxNoteCount);
		for (int i = 0; i < maxNoteCount; i++) {
			aveNoteDur.add(i, (float)0);
			stdDevNoteDur.add(i, (float)0);
		}

		for (int i = 0; i < sngList.getSize(); i++) {
			MPPerformance perf = sngList.Get(i);
			for (int j = 0; j < perf.getNumNotes(); j++) {
				float dur = (float)perf.getNoteDur(j);
				aveNoteDur.set(j, aveNoteDur.get(j) + dur);
				stdDevNoteDur.set(j, stdDevNoteDur.get(j) + dur * dur);
			}
		}
		
		float cnt = sngList.getSize();
		for (int j = 0; j < maxNoteCount; j++) {
			
			float average = aveNoteDur.get(j) / cnt;
			aveNoteDur.set(j, average);
			
			average = average * average;
			average = average * (cnt / (cnt - 1));
			float stddev = stdDevNoteDur.get(j);
			stddev = stddev / (cnt - 1);
			stddev = (float)Math.sqrt(stddev - average);
			stdDevNoteDur.set(j, stddev);
		}		
	}
	
	protected void identifyChords(MPSongList list)
	{
		
	}
	
	protected void identifyChords3(MPSongList list)
	{
		chords = new ArrayList<ChordInfo>(maxNoteCount);
		short values[] = new short[maxNoteCount];
		for (MPPerformance p : list.songs) { 
			for (int i = 0; i < maxNoteCount; i++) {
				if (p.getNoteDur(i) == 0) {
					values[i]++;
				}
			}
		}
		
		for (int i = 0; i < maxNoteCount; i++) {
			if (values[i] > 3) {
				System.out.println(i + " " + aveNoteDur.get(i) + " " + stdDevNoteDur.get(i));
			}
		}
	}
	
	protected void identifyChords2(MPSongList list)
	{
		chords = new ArrayList<ChordInfo>(maxNoteCount);
		
		// build a distribution of the keys struck for each note event
		short values[][] = new short[maxNoteCount][128];
		for (MPPerformance p : list.songs) { 
			for (int i = 0; i < maxNoteCount; i++) {
				short key = (short)p.midi.noteEvents.get(i).getMIDIValue();
				values[i][key]++;
			}
		}
		
		num_performances = list.getSize();
		
		// find the max key for each event
		for (int i = 0; i < maxNoteCount; i++) {
			int max_key_value = 0;
			short max_key = 0;
			for (short key = 0; key < 128; key++) {
				if (values[i][key] > max_key_value) {
					max_key_value = values[i][key];
					max_key = key;
				}
			}
			
			// check for false positive (mistake notes)
			boolean is_chord = false;
			if (max_key_value < .95f * num_performances) {
				is_chord = true;
				System.out.println(i);
			}
			chords.add(i, new ChordInfo(max_key, max_key_value, is_chord));
		}
	}
	
	private void identifyChordBoundaries()
	{
		// post-process chord list; look for chord boundaries
		int i = 0;
		while (i < maxNoteCount) {
			// find beginning of next chord
			while (i < maxNoteCount && !chords.get(i).is_chord) {
				++i;
			}

			// find end of this chord
			if (i < maxNoteCount) {
				float dur = aveNoteDur.get(i);
				int j = i + 1;
				int k = 0;
				int m = i;
				float max_dur = dur;
				while (j < maxNoteCount && chords.get(j).is_chord && k < 3) {
					float dur2 = aveNoteDur.get(j);
					if (dur2 > dur * 3 / 2) {
						m = j;
						break;
					}
					if (dur2 > max_dur) {
						max_dur = dur2;
						m = j;
					}
					++j;
					++k;
				}
				// mark chord boundary
				chords.get(m).boundary = true;
				i = m + 1;
			}
		}
	}
	
	private void adjustChordDurations(MPSongList list)
	{
		// adjust durations so that each note is as if it were not part of chord (durs are deltas between notes)
		for (MPPerformance p : list.songs) {
			int i = 0;
			while (i < maxNoteCount) {
				while (i < maxNoteCount && !chords.get(i).is_chord) {
					++i;
				}
				if (i < maxNoteCount) {
					int j = i;
					while (!chords.get(j).boundary) {
						++j;
					}
					// we have a chord from [i-j]
					long dur = 0;
					for (int k = j; k >= i; k--) {
						dur = p.noteDur.get(k) + dur;
						p.noteDur.set(k, dur);
					}

					i = j + 1;
				}
			}
		}		
	}
	
	private void reorderNotesinChords(MPSongList list) 
	{
		// re-order notes
		for (MPPerformance p : list.songs) {
			int i = 0;
			while (i < maxNoteCount) {
				while (i < maxNoteCount && !chords.get(i).is_chord) {
					++i;
				}
				if (i < maxNoteCount) {
					int j = i;
					while (!chords.get(j).boundary) {
						++j;
					}
					// we have a chord from [i-j]
					for (int k = i; k < j; k++) {
						short key = chords.get(k).key;
						if (p.midi.noteEvents.get(k).getMIDIValue() != key) {
							for (int l = k + 1; l <= j; l++) {
								if ((short)p.midi.noteEvents.get(l).getMIDIValue() == key) {
									p.SwapNotes(l, k);
									break;
								}
							}
						}
					}
					i = j + 1;
				}
			}
			
			/*
			 * weight mistakes???
			float stddev = p.noteStdDeviation;
			for (i = 0; i < maxNoteCount; i++) {
			
				if (chords.get(i).key != p.midi.noteEvents.get(i).getMIDIValue()) {
					long value = p.noteDur.get(i);
					value += stddev;
					p.noteDur.set(i, value);
				}
			}
			*/
		}
		
	}
	
	public void calcChords(MPSongList list)
	{
		identifyChords(list);
		identifyChordBoundaries();
		
		// ok.  we have populated the chords structure.  now traverse the performances and re-order notes and
		// adjust timings to account for chords.  we do this to enhance the correlation analysis.  we shouldn't
		// do this if we are looking at slopes.
		
		adjustChordDurations(list);

		reorderNotesinChords(list);
		
		// update song stats
		for (MPPerformance p : list.songs) {
			p.calcPerfStats();
		}
	}
	
	public void computeStats(MPSongList list)
	{
		if (list.getSize() < 1) {
			return;
		}
		computeSongStats(list);
		computeNoteStats(list);
	}
	
	public void computeAveSlope(MPSongList list, ArrayList<Float> baseline)
	{
		float aveslope = 0;
		int size = Math.min(baseline.size(), maxNoteCount);
		aveNoteSlopeX = new ArrayList<Float>(maxNoteCount);

		if (size <= 1) {
			return;
		}
		
		for (MPPerformance p : list.songs) {
			p.CalcSlope(baseline);
		}
		
		for (int i = 0; i < size; i++) {
			aveslope += baseline.get(i) - aveNoteDur.get(i);
			aveNoteSlopeX.add(i, (float)aveslope);
		}
		for (int i = size; i < maxNoteCount; i++) {
			aveNoteSlopeX.add(i, (float)aveslope);
		}
	}
	
	/*
	
	public int computeMistakes(ArrayList<MPPerformance> list, MPOriginalPerformance orgP)
	{
		if (list.size() <= 1) {
			return 0;
		}
		
		int mistakes = 0;
		for (MPPerformance p : list) {
			int i = p.calcMistakes(orgP);		
			mistakes += i;
			System.out.println(i);
		}
		
		System.out.println("mistakes " + mistakes);
		
		return mistakes;
	}
	
	*/

	/***
	 * computeRepeatSongPercent computes the percentage of songs that we repeated by a given user
	 * 
	 * @return repeat song percent
	 * 
	 */
	public void computeRepeatSongPercent(MPSongList sngList)
	{
		sngList.Sort();

		int repeat_cnt = 0;
		int size = sngList.getSize();
		int i = 0; 
		
		while (i < size - 1) {
			MPPerformance a = sngList.Get(i);
			int j = i + 1;
			boolean repeat = true;
			while (repeat && j < size) {
				MPPerformance b = sngList.Get(j);
				if (a.getUserID() == b.getUserID()) {
					++repeat_cnt;
					++j;
				} else {
					repeat = false;
				}
			}
			i = j;
		}		
		
		repeatSongPercent = (float)repeat_cnt / (float)size;
	}

	
	public void ReportNoteStats(MPOutputStream out, MPOriginalPerformance baseline)
	{		
		out.MPWriteLn("");
		out.MPWrite("Note index");
		for (int j = 0; j < maxNoteCount; j++) {
			out.MPWrite("\t" + j);
		}
		
		/*
		out.MPWriteLn("");
		out.MPWrite("chords");
		for (int j = 0; j < maxNoteCount; j++) {
			String s = "";
			if (chords.get(j).boundary) {
				s = "&";
			} else if (chords.get(j).is_chord) {
				s = "*";
			}
			out.MPWrite("\t" + s);
		}
		
		out.MPWriteLn("");
		for (int j = 0; j < maxNoteCount; j++) {
			// MPNoteEvent.getPitch(
			out.MPWrite("\t" + (chords.get(j).is_chord ? chords.get(j).key : ""));
		}
		out.MPWriteLn("");
		for (int j = 0; j < maxNoteCount; j++) {
			out.MPWrite("\t" + (chords.get(j).is_chord ? chords.get(j).value / (float)num_performances : ""));
		}
		*/
	
		if (baseline != null) {
			out.MPWriteLn("");
			out.MPWrite("pitches");
			for (int j = 0; j < baseline.getNumNotes(); j++) {
				out.MPWrite("\t" + baseline.midi.noteEvents.get(j).getPitch());
			}
	
			out.MPWriteLn("");
			out.MPWrite("Org score");
			for (int j = 0; j < baseline.getNumNotes(); j++) {
				out.MPWrite("\t" + baseline.noteDur.get(j));
			}
		}
	
		out.MPWriteLn("");
		out.MPWrite("Note ave");
		for (int j = 0; j < aveNoteDur.size(); j++) {
			out.MPWrite("\t" + aveNoteDur.get(j));
		}
			
		if (baseline != null) {
			out.MPWriteLn("");
			out.MPWrite("Org - Ave");
			int size = Math.min(baseline.getNumNotes(), aveNoteDur.size());
			for (int j = 0; j < size; j++) {
				out.MPWrite("\t" + (baseline.noteDur.get(j) - aveNoteDur.get(j)));
			}
		}
			
		out.MPWriteLn("");
		out.MPWrite("Note std dev");
		for (int j = 0; j < stdDevNoteDur.size(); j++) {
			out.MPWrite("\t" + stdDevNoteDur.get(j));
		}
		out.MPWriteLn("");

		ReportAveSlope(out);
		out.MPWriteLn("");
	}

	public void ReportAveSlope(MPOutputStream out)
	{
		if (aveNoteSlopeX == null) {
			System.out.println("ReportAveSlope:  Error:  have not calc-ed aveSlope");
			return;
		}
		
		out.MPWrite("Ave note slope");
		for (int j = 0; j < aveNoteSlopeX.size(); j++) {
			out.MPWrite("\t" + aveNoteSlopeX.get(j));
		}
	}

}

