import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.TreeMap;
import java.util.TreeSet;

public class MPUniqueIDs {
	
	public class MPUniqueIDItem implements Comparable<MPUniqueIDItem>  {
		protected int uniqueID;
		protected MPGeoCoord2 loc ;
		
		public MPUniqueIDItem(int id, MPGeoCoord2 coord)
		{
			uniqueID = id;
			loc = coord;
		}
		
		/***
		 * todo:  reimplement compare to geocoord to see how often people travel
		 */
		
		public int compareTo(MPUniqueIDItem inst)
		{
			int result = uniqueID - inst.uniqueID;
			return result;
		}
		
	}
	
	public class MPUniqueIDInfo {
		protected int region_count;
		protected int urban_count;
		protected String region_full_name;
		
		public MPUniqueIDInfo(int regions, int cities, String full_name)
		{
			region_count = regions;
			urban_count = cities;
			region_full_name = full_name;
		}
		
		public int GetRegionCnt()
		{
			return region_count;
		}
		
		public int GetUrbanCount()
		{
			return urban_count;
		}
	}
	
	private TreeSet<MPUniqueIDItem> idTree;
	private TreeMap<Object, String> tmCountryNames;
	
	public MPUniqueIDs()
	{
		idTree = new TreeSet<MPUniqueIDItem>();
		tmCountryNames = new TreeMap<Object, String>();
	}
	
	public void Add(int uniqueID, MPGeoCoord2 coord)
	{
		MPUniqueIDItem newitem = new MPUniqueIDItem(uniqueID, coord);
		idTree.add(newitem);
	}

	private Iterator<MPUniqueIDItem> GetIterator()
	{
		return idTree.iterator();
	}
	
	public void Merge(MPUniqueIDs list2)
	{
		idTree.addAll(list2.idTree);
	}
	
	public void CheckIntegrity()
	{
		ArrayList<MPUniqueIDItem> items = new ArrayList<MPUniqueIDItem>(GetSize());
		
	    Iterator<MPUniqueIDItem> sitr= GetIterator();
	    while (sitr.hasNext()) {
	    	MPUniqueIDItem item = (MPUniqueIDItem)sitr.next();
	    	items.add(item);
	    }
	    
	    Collections.sort(items);
	    for (int i = 0; i < items.size() - 1; i++) {
	    	if (items.get(i).compareTo(items.get(i+1)) == 0) {
	    		System.out.println("Integrity check error:  dups");
	    	}
	    }
	    
	}
	
	public int GetSize()
	{
		return idTree.size();
	}
		
	public MPUniqueIDInfo GetUniqueIDInfoCountByRegion(RIRegionItem region)
	{
		int region_count = 0;
		int urban_count = 0;
		String full_name = null;

	    Iterator<MPUniqueIDItem> sitr= GetIterator();
	    while (sitr.hasNext()) {
	    	MPUniqueIDItem item = (MPUniqueIDItem)sitr.next();
	    	if (region.Contains(item.loc)) {
	    		++region_count;
	    		if (item.loc.isCity()) {
	    			++urban_count;
	    		}
	    		if (full_name == null) {
	    			full_name = region.GetFullName();
	    		}
	    	}
	    }
	    
	    tmCountryNames.put(region.GetRegionName(), full_name);
		MPUniqueIDInfo info = new MPUniqueIDInfo(region_count, urban_count, full_name);
	    return info;
	}
	
	public void OutputAllUniquesGeoRegion(String region)
	{
		MPOutputStream out = new MPOutputStream("regions/" + region + ".txt");
	    Iterator<MPUniqueIDItem> sitr= GetIterator();
	    while (sitr.hasNext()) {
	    	MPUniqueIDItem item = (MPUniqueIDItem)sitr.next();
	    	String itemname = RIGeoRegionItem.GetKey(item.loc);
	    	if (region.matches(itemname)) {
	    		String s = item.loc.countryNameCode + " " + item.loc.administrativeAreaName + " " + item.loc.localityName;
	    		s = s + "\t" + item.loc.n_lat + "," + item.loc.n_long + "\t";
	    		if (item.loc.isCity()) {
	    			s = s + "urban";
	    		} else {
	    			s = s + "rural";
	    		}
	    		out.MPWriteLn(s);
	    	}
	    }
		out.MPClose();
	}
	
	public String getRegionFullName(String region)
	{
		if (tmCountryNames == null) {
			return region;
		}
		String s = (String)tmCountryNames.get(region);
		if (s == null) {
			return region;
		}
		return s;
	}
	
	public void Report(MPOutputStream out)
	{	
	}

}
