import java.util.ArrayList;
import java.util.TreeSet;

/***
 * 
 * @author jeffreysmith
 * 
 * v1.0 November 2012
 * 
 * a basic data structure to store geo coords of a city
 * 
 */

public class RIGeoRegionItem extends RIRegionItem {
	
	// public String city_name = null;
	public float r_lat;
	public float r_long;
	public float r_radius;
	
	// 1 yields 100km/1 or 100 sq km regions
	// 5 yields 100km/5 or 20 sq km regions
	// 10 yields 100km/10 or 10 sq km regions
	// 100 yields 100km/100 or 1 sq km regions
	private static float r_region_size = 100f / MPAnalyzer.GetConfig().geo_region_size;
	private static float r_default_radius = MPAnalyzer.GetConfig().default_region_radius;
	private static float r_region_scale = MPAnalyzer.GetConfig().geo_region_scale;
	
	private static float dRound(float d) 
	{
		d =  Math.round(d * r_region_size) / r_region_size; 
		return d;
	}
	
	private void init(float y_lat, float x_long)
	{
		// round to r_distance
		r_lat = dRound(y_lat);
		r_long = dRound(x_long);
		
		region_name = r_lat + "," + r_long;
		r_radius = r_default_radius;
	}
	
	static public String GetKey(MPGeoCoord2 coord) 
	{
		return dRound(coord.n_lat) + "," + dRound(coord.n_long); 
	}

	public RIGeoRegionItem(MPGeoCoord2 coord)
	{
		super();
		init(coord.n_lat, coord.n_long);
	}
	
	protected RIGeoRegionItem(gPoint pt)
	{
		super();
		r_lat = (float)pt.y;
		r_long = (float)pt.x;
		r_radius = (float)pt.r;
		region_name = r_lat + "," + r_long;
	}
	
	public boolean Contains(MPGeoCoord2 geocoord)
	{
		return Contains(geocoord, r_region_scale);
	}	
	
	public boolean Contains(MPGeoCoord2 geocoord, double scale)
	{
		if (false && geocoord.isCity()) {
			return false;
		}
		
		float dx = r_long - geocoord.n_long;
		float dy = r_lat - geocoord.n_lat;
		double radius = Math.sqrt(dx * dx + dy * dy);
		
		return radius < (r_radius * scale);
	}

	public class gPoint {
		double x;
		double y;
		double r;
		
		public gPoint()
		{
			
		}
		
		public gPoint(double x1, double y1, double r1) 
		{
			x = x1;
			y = y1;
			r = r1;
		}
		
		public boolean Intersects(RIItem i)
		{
			double dx = x - i.x;
			double dy = y - i.y;
			double radius = Math.sqrt(dx * dx + dy * dy);
			return radius < r;
		}
	}
	
	private int uniquesAtPoint(gPoint g)
	{
		TreeSet<Integer> uniques = new TreeSet<Integer>();
		for (RIItem i : perfs) {
			if (g.Intersects(i)) {
				uniques.add(i.id);
			}
		}
		return uniques.size();
	}

	public class gPointLst {
		public ArrayList<gPoint> pts;
		
		// reference to parent perfs
		public ArrayList<RIItem> perfs;
		
		public gPointLst(int size, ArrayList<RIItem> performances)
		{
			perfs = performances;
			pts = new ArrayList<gPoint>(size);
			for (int i = 0; i < size; i++) {
				pts.add(new gPoint());
			}
		}
				
		protected int sumUniques() 
		{
			TreeSet<Integer> uniques = new TreeSet<Integer>();
			for (gPoint g : pts) {
				for (RIItem i : perfs) {
					if (g.Intersects(i)) {
						uniques.add(i.id);
					}
				}
			}
			return uniques.size();
		}

		protected int maxUniques() 
		{
			int max = 0;
			for (gPoint g : pts) {
				int cnt = uniquesAtPoint(g);
				if (cnt > max) {
					max = cnt;
				}
			}
			return max;
		}

		private int minUniques() 
		{
			int min = Integer.MAX_VALUE;
			for (gPoint g : pts) {
				int cnt = uniquesAtPoint(g);
				if (cnt < min) {
					min = cnt;
				}
			}
			return min;
		}
		
		private void setPoints(gPoint origin, double theta, double r)
		{
			for (gPoint p : pts) {
				p.x = origin.x + r * Math.cos(theta);
				p.y = origin.y + r * Math.sin(theta);
				p.r = r;
				theta += Math.PI/2;
			}
		}
		
		private void optimizePts(gPoint origin, double radius, int min_uniques) 
		{
			int max = 0;
			double theta = 0;
			double optimal_theta = theta;
			while (theta < Math.PI * 2) {
				setPoints(origin, theta, radius);
				int cnt = minUniques();
				if (cnt >= min_uniques) {
					optimal_theta = theta;
					break;
				}
				if (cnt > max) {
					max = cnt;
					optimal_theta = theta;
				}
				theta += Math.PI / 12;
			}
			
			// didn't succeeded in finding group with all > minimum; so just use best we found
			if (theta >= Math.PI * 2) {
				setPoints(origin, optimal_theta, radius);
			}
		}
	}


	/*
	 * subdivide region (a circle at origin w/ radius) into regions < max_uniques
	 * 
	 * attempt to make each subdivided region > min_uniques; 
	 * 
	 */
	private void doSubdivide(gPointLst adds, gPoint origin, int min_uniques, int max_uniques, double min_radius)
	{
		// start by breaking circle into four circles
		// zero overlap
		double c_max = Math.sqrt(origin.r * origin.r + origin.r * origin.r) / 2;
		
		// max overlap
		// double c_min = origin.r / 2;
		int uniques = uniquesAtPoint(origin);
		
		// adjust overlap based on ratio of where we are to target
		//double percent = Math.log10((float)uniques / (float)max_uniques);
		//percent = Math.min(percent, 1.0);
		// double c = c_min + (c_max - c_min) * percent;
		double c = c_max;
		gPointLst list = new gPointLst(4, perfs);
		list.optimizePts(origin, c, min_uniques);

		// we may or may not have succeeded in all groups > min_uniques; we proceed nonetheless
		for (gPoint p : list.pts) {
			uniques = uniquesAtPoint(p);
			// check for further subdivision; in this case, sub-divided regions of this will be added
			if (uniques > max_uniques && c <= min_radius) {				
				doSubdivide(adds, p, min_uniques, max_uniques, min_radius);
				
			// this point is not subdivided further, add it; but first grow it until we hit min points inside
			} else if (uniques > 0) {
				/*
				double max_radius = 2 * c / 3;
				while (p.r < max_radius && uniques < min_uniques) {
					p.r = 1.01 * p.r;
					uniques = uniquesAtPoint(p);
				}
				double r_prev;
				do {
					r_prev = p.r;
					p.r = 1.01 * p.r;
					uniques = uniquesAtPoint(p);
				} while (p.r < max_radius && uniques < max_uniques);
				p.r = r_prev;
				*/
				adds.pts.add(p);
			}
		}
	}
	
	private void populateRegion(RIGeoRegionItem gri, gPoint g)
	{
		for (RIItem i : perfs) {
			if (g.Intersects(i)) {
				gri.AddPerformance(i);
			}
		}
	}
	
	@Override
	public boolean Subdivide(ArrayList<RIRegionItem> newitems, int min_uniques, int max_uniques) 
	{
		if (GetNumUniques() <= max_uniques) {
			return false;
		}
		
		gPointLst adds = new gPointLst(0, perfs);
		gPoint origin = new gPoint(r_long, r_lat, r_radius);
		// !!!
		doSubdivide(adds, origin, min_uniques, max_uniques, r_radius / 12);
		
		if (adds.pts.size() < 2) {
			return false;
		}
		
		// System.out.println("subdivide adds " + r_lat + "," + r_long + " " +
		//		adds.pts.size() + " " + GetNumUniques() + " " + adds.sumUniques() + " " + adds.maxUniques());

		// need to update all region items in the adds list using this
		for (gPoint g : adds.pts) {
			RIGeoRegionItem gri = new RIGeoRegionItem(g);
			populateRegion(gri, g);
			newitems.add(gri);
			// System.out.println(" add " + gri.r_lat + "," + gri.r_long + " " + gri.r_radius + " " + gri.GetNumUniques());
		}
		
		return true;
	}
	
	@Override
	public String GetRegionInfo()
	{
		return Float.toString(r_radius);
	}
	

/*
	private boolean get2Points(gPoint org, int max_uniques, gPoint pt1, gPoint pt2)
	{
		double r = org.r;
		double radius_min = .1;
		
		while (r > radius_min) {
			double theta = 0;
			while (theta < 2 * Math.PI) {
				pt1.x = org.x + r / 3 * Math.cos(theta);
				pt1.y = org.y + r / 3 * Math.sin(theta);
				pt1.r = r;
				if (uniquesAtPoint(pt1) < max_uniques) {
					pt2.x = org.x + r / 3 * Math.cos(theta + Math.PI);
					pt2.y = org.y + r / 3 * Math.sin(theta + Math.PI);
					pt2.r = r;
					if (uniquesAtPoint(pt2) < max_uniques) {
						return true;
					}
				}
				theta += Math.PI / 6;
			} // while theta
			r = r * .9;
		} // while radius
		
		return false;
	}
	
	
	private void growRadii(ArrayList<gPoint> adds)
	{
		// find smallest point in terms of uniques; grow its radius
		gPoint index = null;
		int min = Integer.MAX_VALUE;
		for (gPoint g : adds) {
			int cnt = uniquesAtPoint(g);
			if (cnt < min) {
				min = cnt;
				index = g;
			}
		}
		index.r = index.r * 1.1;
	}
	
	private void growToTarget(ArrayList<gPoint> adds, int target)
	{
		while (sumUniques(adds) < target) {
			growRadii(adds);
		}
	}
	
	private boolean get2PointsMin(gPoint org, int min_uniques, gPoint pt1, gPoint pt2)
	{
		double r = org.r / 3;
		double radius_max = org.r * 2;
		
		while (r < radius_max) {
			double theta = 0;
			while (theta < 2 * Math.PI) {
				pt1.x = org.x + r / 2 * Math.cos(theta);
				pt1.y = org.y + r / 2 * Math.sin(theta);
				pt1.r = r;
				if (uniquesAtPoint(pt1) >= min_uniques) {
					pt2.x = org.x + r / 2 * Math.cos(theta + Math.PI);
					pt2.y = org.y + r / 2 * Math.sin(theta + Math.PI);
					pt2.r = r;
					if (uniquesAtPoint(pt2) >= min_uniques) {
						return true;
					}
				}
				theta += Math.PI / 6;
			} // while theta
			r = r * 1.1;
		} // while radius
		
		return false;
	}

	private boolean doSubdivideMax(ArrayList<gPoint> adds, gPoint origin, int min_uniques, int max_uniques)
	{
		boolean result = false;
		gPoint pt1 = new gPoint();
		gPoint pt2 = new gPoint();
		if (get2Points(origin, min_uniques, pt1, pt2)) {
			if (uniquesAtPoint(pt1) > max_uniques) {
				result = doSubdivide(adds, pt1, min_uniques, max_uniques);
				if (result && uniquesAtPoint(pt2) > max_uniques) {
					result = doSubdivide(adds, pt2, min_uniques, max_uniques);
				}
			} else if (uniquesAtPoint(pt2) > max_uniques) {
				result = doSubdivide(adds, pt2, min_uniques, max_uniques);
			} else {
				result = true;
			}
		}
		
		if (result) {
			// add pt1 and pt2
			adds.add(pt1);
			adds.add(pt2);
			return true;
		}

		return false;
	}
*/


}
