
public class RILatitudeRegionItem extends RIRegionItem {

	float n_value;
	static float multiple = 2.5f;

	static boolean latitude = false;
	static boolean western_hemisphere = false;
	static boolean northern_hemisphere = true;
	static boolean is_asia = false;
	static boolean is_europe = false;
	static boolean is_usa = false;
	static boolean is_china = false;

	public RILatitudeRegionItem(MPGeoCoord2 coord) 
	{
		super();

		region_name = GetKey(coord);
		n_value = getValue(coord);
	}

	private static float dRound(float d, float scale) 
	{
		scale = scale * multiple;
		d = Math.round(d / scale);
		d = d * scale;
		return d;
	}
	
	private static float getValue(MPGeoCoord2 coord)
	{
		return getValue(coord, 1.0f);
	}
	
	private static float getValue(MPGeoCoord2 coord, float scale) 
	{
		float d;
		if (latitude) {
			d = dRound(coord.n_lat, scale);
		} else {
			d = dRound(coord.n_long, scale);
		}
		return d;
	}
	
	static String GetKey(MPGeoCoord2 coord)
	{
		String s = Float.toString(getValue(coord));
		return s;
	}
	
	public boolean Contains(MPGeoCoord2 geocoord)
	{
		return Contains(geocoord, 1.0);
	}
	
	public boolean Contains(MPGeoCoord2 geocoord, double scale)
	{
		
		if (northern_hemisphere) {
			if (geocoord.n_lat <= 0) {
				return false;
			}
		}
		
		if (western_hemisphere) {
			if (!geocoord.NorthAmerica() && !geocoord.SouthAmerica()) {
				return false;
			}
		}
		
		if (is_china && !geocoord.isChina()) {
			return false;
		}
		
		if (is_usa && !geocoord.isUS()) {
			return false;
		}
		
		if (is_europe && !geocoord.Europe()) {
			return false;
		}
		
		if (is_asia && !geocoord.Asia()) {
			return false;
		}
		
		if (getValue(geocoord, (float)scale) == n_value) {
			return true;
		}

		return false;
	}

	@Override
	public String GetFullName()
	{
		return region_name;
	}
}
